import sys

import pytest

import skrf

try:
    from skrf.vi.vna import hp
except ImportError:
    pass

if "matplotlib" not in sys.modules:
    pytest.skip(allow_module_level=True)

@pytest.fixture
def mocked_ff(mocker):
    mocker.patch('skrf.vi.vna.hp.hp8510c.HP8510C.__init__', return_value=None)
    mocker.patch('skrf.vi.vna.hp.hp8510c.HP8510C.write')
    mocker.patch('skrf.vi.vna.hp.hp8510c.HP8510C.write_values')
    mocker.patch('skrf.vi.vna.hp.hp8510c.HP8510C.query')
    mocker.patch('skrf.vi.vna.hp.hp8510c.HP8510C.query_values')

    yield hp.HP8510C('TEST')


@pytest.mark.parametrize(
    'param,expected_query,expected_write,query_response,expected_val,write_val',
    [
        ('id', 'OUTPIDEN;', None, 'HP8510C.07.14: Aug 26  1998', 'HP8510C.07.14: Aug 26  1998', None),
        ('freq_start', 'STAR;OUTPACTI;', 'STEP; STAR 100;', '100', 100, 100),
        ('freq_stop', 'STOP;OUTPACTI;', 'STEP; STOP 100;', '100', 100, 100),
        ('is_continuous', 'GROU?', 'CONT;', '"CONTINUAL"', True, True),
    ]
)
def test_params(
    mocker,
    mocked_ff,
    param,
    expected_query,
    expected_write,
    query_response,
    expected_val,
    write_val
):
    mocked_ff._resource = None
    if expected_query is not None:
        mocked_ff.query.return_value = query_response
        test_val = getattr(mocked_ff, param)
        mocked_ff.query.assert_called_once_with(expected_query)
        assert test_val == expected_val

    if expected_write is not None:
        setattr(mocked_ff, param, write_val)
        mocked_ff.write.assert_called_once_with(expected_write)

def test_freq_query(mocker, mocked_ff):
    mocked_ff.query.side_effect = [
        '100', '200', '51'
    ]
    test = mocked_ff.frequency
    assert test == skrf.Frequency(100, 200, 51, unit='hz')

def test_freq_write(mocker, mocked_ff):
    mocked_ff._resource = mocker.MagicMock()
    mocked_ff.min_hz = 100
    mocked_ff.max_hz = 100000
    test_f = skrf.Frequency(100, 200, 51, unit='hz')
    mocked_ff.frequency = test_f
    assert mocked_ff.frequency==test_f

def test_reset(mocker, mocked_ff):
    mocked_ff.reset()
    mocked_ff.write.assert_called_once_with("FACTPRES;")

def test_wait_until_finished(mocker, mocked_ff):
    mocked_ff.wait_until_finished()
    mocked_ff.query.assert_called_once()



import os
import unittest

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_almost_equal

import skrf as rf
from skrf.media import Coaxial


class MediaTestCase(unittest.TestCase):
    """

    """
    def setUp(self):
        """

        """
        self.files_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'qucs_prj'
            )



    def test_line(self):
        """
        """
        fname = os.path.join(self.files_dir,\
                'coaxial.s2p')
        qucs_ntwk = rf.Network(fname)

        a_media = Coaxial(
            frequency = qucs_ntwk.frequency,
            Dint=1e-3, Dout=3e-3, epsilon_r=2.29,
            tan_delta=4e-4, sigma=1./1.68e-8,
            z0_port = 50.
            )
        skrf_ntwk = a_media.line(200e-3,'m')
        # Equal assertion fails if tan_delta or resistivity are non-zero
        #self.assertEqual(qucs_ntwk, skrf_ntwk)
        self.assertTrue(
            max(abs(skrf_ntwk.s_mag[:,1,0] - qucs_ntwk.s_mag[:,1,0])) < 1e-3
            )


    def test_init_from_attenuation_VF_units(self):
        """
        Test the attenuation unit conversions in the Coaxial classmethod
        `from_attenuation_VF_units`.
        """
        # create a dummy Coaxial media for various attenuation and test the
        # resulting alpha values (real part of gamma)
        rng = np.random.default_rng()
        frequency = rf.Frequency(rng.random(), unit='GHz', npoints=1)
        _att = rng.random()
        # dB/m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/m')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att))
        # dB/100m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/100m')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att)/100)
        # dB/feet
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/feet')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att)*rf.meter_2_feet())
        # dB/100m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='dB/100feet')
        assert_almost_equal(coax.gamma.real,  rf.db_2_np(_att)/100*rf.meter_2_feet())
        # Neper/m
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='Np/m')
        assert_almost_equal(coax.gamma.real,  _att)
        # Neper/feet
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=_att, unit='Np/feet')
        assert_almost_equal(coax.gamma.real,  _att*rf.meter_2_feet())

        with self.assertRaises(ValueError):
            coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=np.array([.1, .2]), unit='Np/feet')
        frequency = rf.Frequency(1., 1.1, unit='GHz', npoints=2)
        coax = Coaxial.from_attenuation_VF(frequency=frequency, VF=1, att=np.array([.1, .2]), unit='Np/feet')
        self.assertEqual( coax.frequency.f.shape, (2,))


    def test_init_from_attenuation_VF_array_att(self):
        """
        Test passing array as attenuation in the Coaxial classmethod
        `from_attenuation_VF_units`.
        """
        # create a Coaxial media for frequency-dependent attenuation and
        # test the resulting alpha values (real part of gamma)
        frequency = rf.Frequency(start=1, stop=2, unit='GHz', npoints=101)
        # k0k1k2 attenuation model
        # values taken for HUBER+SUHNER DATA SHEET Coaxial Cable S_10172_B-1
        # attenuation in dB/m for frequency in GHz
        att = 0 + 0.0826*np.sqrt(frequency.f_scaled) + 0.0129*frequency.f_scaled

        coax = Coaxial.from_attenuation_VF(frequency=frequency, att=att, unit='dB/m')
        # check alpha in gamma
        assert_array_almost_equal(rf.db_2_np(att), coax.gamma.real)

        # if the attenuation array length does not match the frequency,
        # should raise a ValueError
        frequency2 = rf.Frequency(start=1, stop=2, unit='GHz', npoints=10)
        with self.assertRaises(ValueError):
            coax = Coaxial.from_attenuation_VF(frequency=frequency2, att=att)

    def test_R(self):
        freq = rf.Frequency(0, 100, 2, unit='GHz')

        rho = 1e-7
        dint = 0.44e-3
        coax = Coaxial(freq, z0_port = 50, Dint = dint, Dout = 1.0e-3,
                       sigma = 1/rho)

        dc_res = rho / (np.pi * (dint/2)**2)

        # Old R calculation valid only when skin depth is much smaller
        # then inner conductor radius
        with pytest.warns(RuntimeWarning, match="divide by zero"):
            R_simple = coax.Rs/(2*np.pi)*(1/coax.a + 1/coax.b)

            self.assertTrue(abs(1 - coax.R[0]/dc_res) < 1e-2)
            self.assertTrue(abs(1 - coax.R[1]/R_simple[1]) < 1e-2)



"""
media (:mod:`skrf.media.media`)
========================================

Media class.

.. autosummary::
    :toctree: generated/

    Media
    DefinedGammaZ0

"""
from __future__ import annotations

import re
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy as copy
from numbers import Number
from pathlib import Path

import numpy as np
from numpy import any, array, gradient, imag, ones, real
from scipy import stats
from scipy.constants import c

from .. import mathFunctions as mf
from ..constants import S_DEF_DEFAULT, ZERO, NumberLike, to_meters
from ..frequency import Frequency
from ..network import Network, connect, impedance_mismatch, innerconnect


class Media(ABC):
    """
    Abstract Base Class for a single mode on a transmission line media.


    This class init's with `frequency` and `z0_port` (the port impedance);
    attributes shared by all media. Methods defined here make use of the
    properties :

    * `gamma` - (complex) media propagation constant
    * `z0` - (complex) media characteristic impedance

    Which define the properties of a specific media. Any sub-class of Media
    must implement these properties. `gamma` and `z0` should return
    complex arrays of the same length as `frequency`. `gamma` must
    follow the convention:

    * positive real(gamma) = attenuation
    * positive imag(gamma) = forward propagation

    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object or None
        frequency band of this transmission line medium.
        Defaults to None, which produces a 1-10ghz band with 101 points.
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    z0_override : number, array-like, or None
        `z0_override` override the characteristic impedance for the media.
        If `z0_override` is not None, the networks generated by the media have
        their characteristic impedance `z0` overrided by `z0_override`.
        (Default is None)
    z0 : number, array-like, or None
        deprecated parameter, alias to `z0_override` if `z0_override` is None.
        Emmit a deprecation warning.


    Notes
    -----
    Full-wave simulations results can be networks with complex and
    frequency-dependant `z0` port impedances.
    However, the regular touchstone (s-parameters) files generated by most VNA
    measurements are normalized to a single real value, e. g. 1. or 50 ohm.
    Manually setting the `z0_port` to a single real value enforce the second
    behaviour. The default (`z0_port = None`) is the first behaviour.
    """

    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0_override: NumberLike | None = None,
                 z0: NumberLike | None = None):
        if frequency is None:
            frequency = Frequency(1,10,101,'GHz')

        if z0 is not None:
            # warns of deprecation
            warnings.warn(
                'Use of `z0` in Media initialization is deprecated.\n'
                'If `z0_override` is None, '
                'then `z0_override` will take z0 value.\n'
                '`z0` will be removed of Media initialization  in version 1.0',
              DeprecationWarning, stacklevel = 2)
            if z0_override is None:
                z0_override = z0

        self.frequency = frequency.copy()
        self.z0_port = z0_port
        self.z0_override = z0_override

    def mode(self,  **kw) -> Media:
        r"""
        Create another mode in this medium.

        Convenient way to return a copy this Media object
        with eventually different properties.

        Parameters
        ----------
        \*\*kwargs : keyword arguments passed to the copy

        Returns
        -------
        copy : :class:`Media`
            A copy of this Media object with \*\*kwargs attribute
        """
        out = copy(self)
        for k in kw:
            setattr(self, k, kw[k])
        return out

    def copy(self) -> Media:
        """
        Copy of this Media object.

        Returns
        -------
        copy : :class:`Media`
            A copy of this Media object
        """
        return copy(self)

    def __eq__(self,other):
        """
        Test for numerical equality (up to :data:`~skrf.constants.ZERO`).
        """

        if self.frequency != other.frequency:
            return False

        if max(abs(self.z0 - other.z0)) > ZERO:
            return False

        if max(abs(self.gamma - other.gamma)) > ZERO:
            return False

        if self.z0_port is None and other.z0_port is not None:
            if max(abs(self.z0 - other.z0_port)) > ZERO:
                return False

        return True

    def __len__(self) -> int:
        """
        Length of frequency axis.
        """
        return len(self.frequency)

    @property
    def npoints(self) -> int:
        """
        Number of points of the frequency axis.

        Returns
        -------
        npoints : int
            Number of points of the frequency axis.
        """
        return self.frequency.npoints

    @npoints.setter
    def npoints(self, val):
        self.frequency.npoints = val

    @property
    def z0_port(self) -> np.ndarray:
        """
        Port Impedance.

        Returns
        -------
        z0_port : :class:`numpy.ndarray`
        """
        if self._z0_port is None:
            return self._z0_port
        return self._z0_port*ones(len(self))

    @z0_port.setter
    def z0_port(self, val):
        self._z0_port = val

    @property
    def z0_override(self) -> np.ndarray:
        """
        Port Impedance.

        Returns
        -------
        z0_port : :class:`numpy.ndarray`
        """
        if self._z0_override is None:
            return self._z0_override
        return self._z0_override*ones(len(self))

    @z0_override.setter
    def z0_override(self, val):
        self._z0_override = val

    @property
    @abstractmethod
    def gamma(self):
        r"""
        Propagation constant.

        In skrf, defined as :math:`\gamma = \alpha + j \beta`.

        Returns
        -------
        gamma : :class:`numpy.ndarray`
            complex propagation constant for this media

        Notes
        -----
        `gamma` must adhere to the following convention:

         * positive real(gamma) = attenuation
         * positive imag(gamma) = forward propagation
        """
        return None


    @property
    def alpha(self) -> np.ndarray:
        """
        Real (attenuation) component of gamma.

        Returns
        -------
        alpha : :class:`numpy.ndarray`
        """
        return real(self.gamma)

    @property
    def beta(self) -> np.ndarray:
        """
        Imaginary (propagating) component of gamma.

        Returns
        -------
        beta : :class:`numpy.ndarray`
        """
        return imag(self.gamma)

    @property
    @abstractmethod
    def z0_characteristic(self):
        """
        Characteristic Impedance, :math:`z_0`.
        This abstract method has to be defined in the Media Class.

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        return None

    @property
    def z0(self):
        """
        Return Characteristic Impedance `z0_characteristic`.
        If `z0_override` is not None, it is returned instead.

        Returns
        -------
        z0 : np.ndarray
            z0_characteristic or z0_override in units of ohms
        """
        if self.z0_override is None:
            return self.z0_characteristic
        else:
            return self.z0_override

    # left for backward compatibility
    @property
    def Z0(self):
        """
        Characteristic Impedance

        Returns
        -------
        z0 : np.ndarray
            Characteristic Impedance in units of ohms
        """
        # warns of deprecation
        warnings.warn(
            'Use of `Z0` (uppercase) is deprecated and return `z0`.\n'
            'Use characteristic impedance `z0` (lowercase) instead.\n'
            '`Z0` will be removed in version 1.0.',
          DeprecationWarning, stacklevel = 2)
        return self.z0


    @property
    def v_p(self) -> np.ndarray:
        r"""
        Complex phase velocity (in m/s).

        .. math::
            j \cdot \omega / \gamma

        Notes
        -----
        The `j` is used so that real phase velocity corresponds to
        propagation

        where:

        * :math:`\omega` is angular frequency (rad/s),
        * :math:`\gamma` is complex propagation constant (rad/m)

        Returns
        -------
        v_p : :class:`numpy.ndarray`

        See Also
        --------
        propagation_constant
        gamma

        """
        return 1j*(self.frequency.w/self.gamma)


    @property
    def v_g(self):
        r"""
        Complex group velocity (in m/s).

        .. math::
            j \cdot d \omega / d \gamma


        where:

        * :math:`\omega` is angular frequency (rad/s),
        * :math:`\gamma` is complex propagation constant (rad/m)

        Notes
        -----
        the `j` is used to make propagation real, this is needed because
        skrf defined the gamma as :math:`\gamma= \alpha +j\beta`.

        Returns
        -------
        v_g : :class:`numpy.ndarray`

        References
        ----------
        https://en.wikipedia.org/wiki/Group_velocity

        See Also
        --------
        propagation_constant
        v_p
        gamma
        """
        dw = self.frequency.dw
        dk = gradient(self.gamma)

        return dw/dk


    def get_array_of(self, x):
        try:
            if len(x)!= len(self):
                # we have to make a decision
                pass
        except(TypeError):
            y = x* ones(len(self))

        return y

    ## Other Functions
    def theta_2_d(self, theta: NumberLike, deg:bool = True, bc: bool = True) -> NumberLike:
        r"""
        Convert electrical length to physical distance.

        The electrical length is given by :math:`d=\theta/\beta`.

        The given electrical length can be given either at the center frequency
        or on the entire band depending of the parameter `bc`.

        Parameters
        ----------
        theta : number
            electrical length, at band center (see deg for unit)
        deg : Boolean, optional
            is theta in degrees?
            Default is True (theta is assumed in degrees)
        bc : bool, optional.
            evaluate only at band center, or across the entire band?
            Default is True (evaluation assumed at band center)

        Returns
        --------
        d : number, array-like
            physical distance in meters

        """
        if deg:
            theta = mf.degree_2_radian(theta)

        gamma = self.gamma
        if bc:
                return 1.0*theta/np.imag(gamma[int(gamma.size/2)])
        else:
                return 1.0*theta/np.imag(gamma)

    def electrical_length(self, d: NumberLike, deg: bool = False) -> NumberLike:
        r"""
        Calculate the complex electrical length for a given distance.

        Electrical length is given by :math:`\theta=\gamma d`.

        Parameters
        ----------
        d: number or array-like
            delay distance, in meters
        deg: Boolean, optional
            return electrical length in deg?
            Default is False (returns electrical length in radians)

        Returns
        -------
        theta: number or array-like
            complex electrical length in radians or degrees, depending on
            value of deg.
        """
        gamma = self.gamma

        if not deg:
            return gamma*d
        else:
            return  mf.radian_2_degree(gamma*d)

    ## Network creation

    # lumped elements
    def match(self, nports: int = 1, z0: NumberLike | None = None,
              z0_norm: bool = False, **kwargs) -> Network:
        r"""
        Perfect matched load (:math:`\Gamma_0 = 0`).

        Parameters
        ----------
        nports : int
            number of ports
        z0 : number, or array-like or None
            port impedance. Default is
            None, in which case the Media's :attr:`z0` is used.
            This sets the resultant Network's
            :attr:`~skrf.network.Network.z0`.
        z0_norm : bool
            is z0 normalized to this media's `z0`?
        \*\*kwargs : key word arguments
            passed to :class:`~skrf.network.Network` initializer

        Returns
        -------
        match : :class:`~skrf.network.Network` object
            a n-port match

        Examples
        --------
        >>> my_match = my_media.match(2,z0 = 50, name='Super Awesome Match')

        """
        result = Network(**kwargs)
        result.frequency = self.frequency
        result.s = np.zeros((self.frequency.npoints, nports, nports), dtype=complex)
        result.port_modes = np.array(["S"] * result.nports)
        if z0 is None:
            if self.z0_port is None:
                z0 = self.z0
            else:
                z0 = self.z0_port
        else:
            if isinstance(z0, str):
                z0 = parse_z0(z0)*self.z0
            if z0_norm:
                z0 = z0*self.z0

        result.z0 = z0
        return result

    def load(self, Gamma0: NumberLike, nports: int = 1,
             z0: NumberLike | None = None, **kwargs) -> Network:
        r"""
        Load of given reflection coefficient.

        Parameters
        ----------
        Gamma0 : number, array-like
            Reflection coefficient of load (linear, not in db). If its
            an array it must be of shape: `kxnxn`, where k is number of frequency
            points in media, and n is `nports`
        nports : int
            number of ports
        z0 : number, or array-like or None
            port impedance. Default is
            None, in which case the Media's :attr:`z0` is used.
            This sets the resultant Network's
            :attr:`~skrf.network.Network.z0`.
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        load : :class:`~skrf.network.Network` object
            n-port load, where  S = Gamma0*eye(...)

        See Also
        --------
        match
        open
        short
        """
        result = self.match(nports, z0 = z0, **kwargs)
        result.s = np.array(Gamma0).reshape(-1, 1, 1) * \
            np.eye(nports, dtype=complex).reshape((-1, nports, nports)).\
            repeat(self.frequency.npoints, 0)
        result.port_modes = np.array(["S"] * result.nports)
        return result

    def short(self, nports: int = 1,
              z0: NumberLike | None = None, **kwargs) -> Network:
        r"""
        Short (:math:`\Gamma_0 = -1`)

        For s_def = 'power' (:math:`-Z_{ref}^*/Z_{ref}`)

        Parameters
        ----------
        nports : int
            number of ports
        \*\*kwargs : key word arguments passed to :func:`load`.
        z0 : number, or array-like or None
            port impedance. Default is
            None, in which case the Media's :attr:`z0` is used.
            This sets the resultant Network's
            :attr:`~skrf.network.Network.z0`.
        Returns
        -------
        match : :class:`~skrf.network.Network` object
            a n-port short circuit

        Notes
        -----
        This calls ::

            load(-1.0, nports, **kwargs)

        See Also
        --------
        match
        open
        load
        """
        if z0 is None:
            if self.z0_port is None:
                z0 = self.z0
            else:
                z0 = self.z0_port
        s_short = -1
        # Powerwave short is not necessarily -1
        if kwargs.get('s_def', S_DEF_DEFAULT) == 'power':
            s_short = -np.conjugate(z0) / z0
        return self.load(s_short, nports, z0 = z0, **kwargs)

    def open(self, nports: int = 1, **kwargs) -> Network:
        r"""
        Open (:math:`\Gamma_0 = 1`).

        Parameters
        ----------
        nports : int
            number of ports
        \*\*kwargs : key word arguments passed to :func:`load`

        Returns
        -------
        match : :class:`~skrf.network.Network` object
            a n-port open circuit

        Notes
        -----
        This calls ::

            load(1.0, nports, **kwargs)

        See Also
        --------
        match
        load
        short
        """

        return self.load(1.0, nports, **kwargs)

    def resistor(self, R: NumberLike, *args, **kwargs) -> Network:
        r"""
        Resistor.

        Parameters
        ----------
        R : number, array
            Resistance , in Ohms. If this is an array, must be of
            same length as frequency vector.
        \*args, \*\*kwargs : arguments, key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        resistor : a 2-port :class:`~skrf.network.Network`

        See Also
        --------
        match
        short
        open
        load
        capacitor
        inductor
        """
        s_def = kwargs.pop('s_def', S_DEF_DEFAULT)
        result = self.match(nports=2, s_def='power', **kwargs)
        s = np.zeros(shape=result.s.shape, dtype=complex)
        R = np.array(R)
        # Convert Y-parameters resistor to S-parameters in power wave to accommodate any R value.
        # y[:, 0, 0] = 1.0 / R
        # y[:, 1, 1] = 1.0 / R
        # y[:, 0, 1] = -1.0 / R
        # y[:, 1, 0] = -1.0 / R
        z0_0, z0_1 = result.z0[:, 0], result.z0[:, 1]
        denom = R + (z0_0 + z0_1)
        s[:, 0, 0] = (R - z0_0.conj() + z0_1) / denom
        s[:, 1, 1] = (R + z0_0 - z0_1.conj()) / denom
        s[:, 0, 1] = 2 * (z0_0.real * z0_1.real)**0.5 / denom
        s[:, 1, 0] = 2 * (z0_0.real * z0_1.real)**0.5 / denom
        result.s = s

        # Renormalize into s_def if required
        if s_def != 'power':
            result.renormalize(z_new=result.z0, s_def=s_def)

        return result

    def capacitor(self, C: NumberLike, **kwargs) -> Network:
        r"""
        Capacitor.

        Parameters
        ----------
        C : number, array
            Capacitance, in Farads. If this is an array, must be of
            same length as frequency vector.
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        capacitor : a 2-port :class:`~skrf.network.Network`

        See Also
        --------
        match
        short
        open
        load
        resistor
        inductor
        """
        s_def = kwargs.pop('s_def', S_DEF_DEFAULT)
        result = self.match(nports=2, s_def='power', **kwargs)
        w = self.frequency.w
        s = np.zeros(shape=result.s.shape, dtype=complex)
        C = np.array(C)
        # Convert Y-parameters capacitor to S-parameters in power wave to accommodate any C value.
        # y[:, 0, 0] = 1j * w * C
        # y[:, 1, 1] = 1j * w * C
        # y[:, 0, 1] = -1j * w * C
        # y[:, 1, 0] = -1j * w * C
        z0_0, z0_1 = result.z0[:, 0], result.z0[:, 1]
        denom = 1.0 + 1j * w * C * (z0_0 + z0_1)
        s[:, 0, 0] = (1.0 - 1j * w * C * (z0_0.conj() - z0_1) ) / denom
        s[:, 1, 1] = (1.0 - 1j * w * C * (z0_1.conj() - z0_0) ) / denom
        s[:, 0, 1] = (2j * w * C * (z0_0.real * z0_1.real)**0.5) / denom
        s[:, 1, 0] = (2j * w * C * (z0_0.real * z0_1.real)**0.5) / denom
        result.s = s

        # Renormalize into s_def if required
        if s_def != 'power':
            result.renormalize(z_new=result.z0, s_def=s_def)

        return result

    def inductor(self, L: NumberLike, **kwargs) -> Network:
        r"""
        Inductor.

        Parameters
        ----------
        L : number, array
            Inductance, in Henries. If this is an array, must be of
            same length as frequency vector.
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        inductor : a 2-port :class:`~skrf.network.Network`

        See Also
        --------
        match
        short
        open
        load
        capacitor
        resistor
        """
        s_def = kwargs.pop('s_def', S_DEF_DEFAULT)
        result = self.match(nports=2, s_def='power', **kwargs)
        w = self.frequency.w
        s = np.zeros(shape=result.s.shape, dtype=complex)
        L = np.array(L)
        # Convert Y-parameters inductor to S-parameters in power wave to accommodate any L value.
        # y[:, 0, 0] = 1.0 / (1j * w * L)
        # y[:, 1, 1] = 1.0 / (1j * w * L)
        # y[:, 0, 1] = -1.0 / (1j * w * L)
        # y[:, 1, 0] = -1.0 / (1j * w * L)
        z0_0, z0_1 = result.z0[:, 0], result.z0[:, 1]
        denom = (1j * w * L) + (z0_0 + z0_1)
        s[:, 0, 0] = (1j * w * L - z0_0.conj() + z0_1) / denom
        s[:, 1, 1] = (1j * w * L + z0_0 - z0_1.conj()) / denom
        s[:, 0, 1] = 2 * (z0_0.real * z0_1.real)**0.5 / denom
        s[:, 1, 0] = 2 * (z0_0.real * z0_1.real)**0.5 / denom
        result.s = s

        # Renormalize into s_def if required
        if s_def != 'power':
            result.renormalize(z_new=result.z0, s_def=s_def)

        return result

    def impedance_mismatch(self, z1: NumberLike, z2: NumberLike, **kwargs) -> Network:
        r"""
        Two-port network for an impedance mismatch.

        Parameters
        ----------
        z1 : number, or array-like
            complex impedance of port 1
        z2 : number, or array-like
            complex impedance of port 2
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        missmatch : :class:`~skrf.network.Network` object
            a 2-port network representing the impedance mismatch

        Notes
        -----
        If z1 and z2 are arrays, they must be of same length
        as the :attr:`Media.frequency.npoints`

        See Also
        --------
        match
        short
        open
        load
        capacitor
        inductor
        resistor
        """
        result = self.match(nports=2, **kwargs)
        s_def = kwargs.get('s_def', S_DEF_DEFAULT)
        z1 = np.array(z1)
        z2 = np.array(z2)
        mismatch = np.broadcast_to(impedance_mismatch(z1, z2, s_def), result.s.shape)
        result.s = mismatch
        return result


    # splitter/couplers
    def tee(self, **kwargs) -> Network:
        r"""
        Ideal, lossless tee. (3-port splitter).

        Parameters
        ----------
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        tee : :class:`~skrf.network.Network` object
            a 3-port splitter

        See Also
        ----------
        splitter : this just calls splitter(3)
        match : called to create a 'blank' network
        """
        return self.splitter(3, **kwargs)

    def splitter(self, nports: int, **kwargs) -> Network:
        r"""
        Ideal, lossless n-way splitter.

        The port impedances can be mismatched and the power is split
        accordingly.

        For n > 2, the splitter is not matched because the power wave entering
        one port meet the equivalent impedance of the other ports in parallel.

        .. math::
            S_{ii} = \frac{Y_i - \sum_{j\neq i} Y_j}{Y_i + \sum_{j\neq i} Y_j}
            = \frac{2Y_i}{\sum_{j=1...n} Y_j} - 1

        The remaining power is split between the other ports depending their
        impedances.

        .. math::
            \sum_{j=1...n} S_{ij}^2 = 1 - S_{ii}^2

        .. math::
            S_{ij} = \frac{2\sqrt{Y_i \cdot Y_j}}{\sum_{k=1...n} Y_{k}}

        Parameters
        ----------
        nports : int
                number of ports
        \*\*kwargs : key word arguments
                passed to :func:`match`, which is called initially to create a
                'blank' network.

        Returns
        -------
        tee : :class:`~skrf.network.Network` object
                a n-port splitter

        See Also
        --------
        match : called to create a 'blank' network
        """
        result = self.match(nports, **kwargs)

        y0s = np.array(1./result.z0)
        y_k = y0s.sum(axis=1)
        s = np.zeros((self.frequency.npoints, nports, nports),
                      dtype='complex')
        s = 2 *np.sqrt(np.einsum('ki,kj->kij', y0s, y0s)) / y_k[:, None, None]
        np.einsum('kii->ki', s)[:] -= 1  # Sii
        result.s =  s
        return result


    # transmission line

    def to_meters(self, d: NumberLike, unit: str = 'deg') -> NumberLike:
        """
        Translate various units of distance into meters.

        This is a method of media to allow for electrical lengths as
        inputs. For dispersive media, mean group velocity is used to
        translate time-based units to distance.

        Parameters
        ----------
        d : number or array-like
            the value
        unit : str
            the unit to that x is in:
            ['deg','rad','m','cm','um','in','mil','s','us','ns','ps']

        Returns
        -------
        d_m : number, array-like
            d in meters

        See Also
        --------
        skrf.constants.to_meters
        """
        unit = unit.lower()
        #import pdb;pdb.set_trace()

        d_dict ={'deg':self.theta_2_d(d,deg=True),
                 'rad':self.theta_2_d(d,deg=False),
                 }

        if unit in d_dict:
            return d_dict[unit]
        else:
            # mean group velocity is used to translate time-based
            # units to distance
            if 's' in unit:
                # they are specifying  a time unit so calculate
                # the group velocity. (note this fails for media of
                # too little points, as it uses gradient)
                v_g = -self.v_g.imag.mean()
            else:
                v_g = c
            return to_meters(d=d,unit=unit, v_g=v_g)

    def thru(self,
             z0: NumberLike | None = None,
             **kwargs) -> Network:
        r"""
        Matched transmission line of length 0.

        Parameters
        ----------
        z0 : number, or array-like or None
            port impedance. Default is
            None, in which case the Media's :attr:`z0` is used.
            This sets the resultant Network's
            :attr:`~skrf.network.Network.z0`.
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        thru : :class:`~skrf.network.Network` object
            matched transmission line of 0 length

        See Also
        --------
        line : this just calls line(0)
        open, short, match
        """
        if z0 is None:
            if self.z0_port is None:
                z0 = self.z0
            else:
                z0 = self.z0_port
        return self.line(0, z0 = z0, **kwargs)

    def line(self, d: NumberLike, unit: str = 'deg',
             z0: NumberLike | str | None = None, embed: bool = False, **kwargs) -> Network:
        r"""
        Transmission line of a given length and impedance.

        The units of `length` are interpreted according to the value
        of `unit`. If `z0` is not None, then a line specified impedance
        is produced. if `embed` is also True, then the line is embedded
        in this media's z0 environment, creating a mismatched line.

        Parameters
        ----------
        d : number
                the length of transmission line (see unit argument)
        unit : ['deg','rad','m','cm','um','in','mil','s','us','ns','ps']
                the units of d.  See :func:`to_meters`, for details
        z0 : number, string, or array-like or None
            the characteristic impedance of the line, if different
            from self.z0. To set z0 in terms of normalized impedance,
            pass a string, like `z0='1+.2j'`
        embed : bool
            deprecated parameter that only emmit a warning if set to true.
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        line : :class:`~skrf.network.Network` object
            matched transmission line of given length

        Examples
        --------
        >>> my_media.line(1, 'mm', z0=100)
        >>> my_media.line(90, 'deg', z0='2') # set z0 as normalized impedance

        """
        if embed:
            # warns of deprecation
            warnings.warn(
                'Use of `embed` is deprecated and has no more effect.\n'
                'Lines are renormalized to the media port impedance `z0_port`'
                'if it is not None, otherwise they use characterisic'
                'impedance `z0`.\n'
                '`embed` will be removed in version 1.0',
              DeprecationWarning, stacklevel = 2)

        if isinstance(z0, str):
            z0 = parse_z0(z0) * self.z0

        if z0 is None:
            z0 = self.z0

        s_def = kwargs.pop('s_def', S_DEF_DEFAULT)

        # The use of either traveling or pseudo waves s-parameters definition
        # is required here.
        # The definition of the reflection coefficient for power waves has
        # conjugation.
        result = self.match(nports=2, z0 = z0, s_def='traveling', **kwargs)

        theta = self.electrical_length(self.to_meters(d=d, unit=unit))

        s11 = np.zeros(self.frequency.npoints, dtype=complex)
        s21 = np.exp(-1*theta)
        result.s = \
                np.array([[s11, s21],[s21,s11]]).transpose().reshape(-1,2,2)

        # renormalize (or embed) into z0_port if required
        if self.z0_port is not None:
            result.renormalize(self.z0_port)
        result.renormalize(result.z0, s_def=s_def)

        return result

    def line_floating(self, d: NumberLike, unit: str = 'deg',
                z0: NumberLike | str | None = None, **kwargs) -> Network:
        r"""
        Floating transmission line of a given length and impedance.

        This method returns a transmission line with floating shields.
        This is a four-port network, as opposed to the two-port _line_ method.
        Ports 1 and 3 are one side of the line, and electrically connect to
        ports 2 and 4 on the other side, respectively.

        The units of `length` are interpreted according to the value
        of `unit`. If `z0` is not None, then a line specified impedance
        is produced.

        Parameters
        ----------
        d : number
                the length of transmission line (see unit argument)
        unit : ['deg','rad','m','cm','um','in','mil','s','us','ns','ps']
                the units of d.  See :func:`to_meters`, for details
        z0 : number, string, or array-like or None
            the characteristic impedance of the line, if different
            from `media.z0`. To set z0 in terms of normalized impedance,
            pass a string, like `z0='1+.2j'`
        \*\*kwargs : key word arguments
            passed to :func:`match`, which is called initially to create a
            'blank' network.

        Returns
        -------
        line_floating : :class:`~skrf.network.Network` object
            matched, floating transmission line of given length

        Examples
        --------
        >>> my_media.line_floating(1, 'mm', z0=100)
        >>> my_media.line_floating(90, 'deg', z0='2') # set z0 as normalized impedance

        """
        if isinstance(z0, str):
            z0 = self.parse_z0(z0) * self.z0

        if z0 is None:
            z0 = self.z0

        s_def = kwargs.pop('s_def', S_DEF_DEFAULT)

        # The use of either traveling or pseudo waves s-parameters definition
        # is required here.
        # The definition of the reflection coefficient for power waves has
        # conjugation.
        result = self.match(nports=4, z0=z0, s_def='traveling', **kwargs)

        t = self.electrical_length(self.to_meters(d=d, unit=unit))

        denom = -1 + 9*np.exp(2*t)
        s11 = (1 + 3*np.exp(2*t)) / denom
        s12 = 4*np.exp(t) / denom
        s13 = (-2 + 6*np.exp(2*t)) / denom
        s14 = -s12

        result.s = np.array([[s11, s12, s13, s14],
                             [s12, s11, s14, s13],
                             [s13, s14, s11, s12],
                             [s14, s13, s12, s11]]).transpose(2, 0, 1)

        # renormalize (or embed) into z0_port if required
        if self.z0_port is not None:
            result.renormalize(self.z0_port)
        result.renormalize(result.z0, s_def=s_def)

        return result


    def delay_load(self, Gamma0: NumberLike, d: Number, unit: str = 'deg', **kwargs) -> Network:
        r"""
        Delayed load.

        A load with reflection coefficient `Gamma0` at the end of a
        matched line of length `d`.

        Parameters
        ----------
        Gamma0 : number, array-like
            reflection coefficient of load (not in dB)
        d : number
            the length of transmission line (see unit argument)
        unit : ['deg','rad','m','cm','um','in','mil','s','us','ns','ps']
            the units of d.  See :func:`to_meters`, for details
        \*\*kwargs : key word arguments
            passed to :func:`line` and :func:`load`

        Returns
        -------
        delay_load : :class:`~skrf.network.Network` object
                a delayed load

        Examples
        ----------
        >>> my_media.delay_load(-.5, 90, 'deg', Z0=50)

        Notes
        -----
        This calls ::

            line(d, unit, **kwargs) ** load(Gamma0, **kwargs)

        See Also
        --------
        line : creates the network for line
        load : creates the network for the load
        delay_short
        delay_open
        """
        return self.line(d=d, unit=unit, **kwargs) ** self.load(Gamma0=Gamma0,
                                                                **kwargs)

    def delay_short(self, d: Number, unit: str = 'deg', **kwargs) -> Network:
        r"""
        Delayed Short.

        A transmission line of given length terminated with a short.

        Parameters
        ----------
        d : number
            the length of transmission line (see unit argument)
        unit : ['deg','rad','m','cm','um','in','mil','s','us','ns','ps']
            the units of d.  See :func:`to_meters`, for details
        \*\*kwargs : key word arguments passed to :func:`delay_load`.

        Returns
        -------
        delay_short : :class:`~skrf.network.Network` object
                a delayed short

        Notes
        -----
        This calls::

                delay_load(Gamma0=-1.0, d=d, unit=unit, **kwargs)

        See Also
        --------
        delay_load
        delay_open
        """
        return self.delay_load(Gamma0=-1.0, d=d, unit=unit, **kwargs)

    def delay_open(self, d: Number, unit: str = 'deg', **kwargs) -> Network:
        r"""
        Delayed open transmission line.

        Parameters
        ----------
        d : number
            the length of transmission line (see unit argument)
        unit : ['deg','rad','m','cm','um','in','mil','s','us','ns','ps']
            the units of d.  See :func:`to_meters`, for details
        \*\*kwargs : key word arguments passed to :func:`delay_load`

        Returns
        -------
        delay_open : :class:`~skrf.network.Network` object
            a delayed open

        Notes
        -----
        This calls::

                delay_load(Gamma0=1.0, d=d, unit=unit, **kwargs)

        See Also
        --------
        delay_load
        delay_short
        """
        return self.delay_load(Gamma0=1.0, d=d, unit=unit, **kwargs)

    def shunt(self, ntwk: Network, **kwargs) -> Network:
        r"""
        Shunts a :class:`~skrf.network.Network`.

        This creates a :func:`tee` and connects
        `ntwk` to port 1, and returns the result.

        Parameters
        ----------
        ntwk : :class:`~skrf.network.Network` object
        \*\*kwargs : keyword arguments passed to :func:`tee`

        Returns
        -------
        shunted_ntwk : :class:`~skrf.network.Network` object
                a shunted a ntwk. The resultant shunted_ntwk will have
                (2 + ntwk.number_of_ports -1) ports.

        See Also
        --------
        shunt_delay_load
        shunt_delay_open
        shunt_delay_short
        shunt_resistor
        shunt_capacitor
        shunt_inductor
        """
        return connect(self.tee(**kwargs), 1, ntwk, 0)

    def shunt_delay_load(self, *args, **kwargs) -> Network:
        r"""
        Shunted delayed load.

        Parameters
        ----------
        \*args,\*\*kwargs : arguments, keyword arguments
            passed to func:`delay_load`

        Returns
        --------
        shunt_delay_load : :class:`~skrf.network.Network` object
            a shunted delayed load (2-port)

        Notes
        -----
        This calls::

                shunt(delay_load(*args, **kwargs))

        See Also
        --------
        shunt
        shunt_delay_open
        shunt_delay_short
        shunt_resistor
        shunt_capacitor
        shunt_inductor
        """
        return self.shunt(self.delay_load(*args, **kwargs), **kwargs)

    def shunt_delay_open(self,*args,**kwargs) -> Network:
        r"""
        Shunted delayed open.

        Parameters
        ----------
        \*args,\*\*kwargs : arguments, keyword arguments
            passed to func:`delay_open`

        Returns
        -------
        shunt_delay_open : :class:`~skrf.network.Network` object
            shunted delayed open (2-port)

        Notes
        -----
        This calls::

                shunt(delay_open(*args, **kwargs))

        See Also
        --------
        shunt
        shunt_delay_load
        shunt_delay_short
        shunt_resistor
        shunt_capacitor
        shunt_inductor
        """
        return self.shunt(self.delay_open(*args, **kwargs), **kwargs)

    def shunt_delay_short(self, *args, **kwargs) -> Network:
        r"""
        Shunted delayed short.

        Parameters
        ----------
        \*args,\*\*kwargs : arguments, keyword arguments
                passed to func:`delay_short`

        Returns
        -------
        shunt_delay_load : :class:`~skrf.network.Network` object
                shunted delayed open (2-port)

        Notes
        -----
        This calls::

                shunt(delay_short(*args, **kwargs))

        See Also
        --------
        shunt
        shunt_delay_load
        shunt_delay_open
        shunt_capacitor
        shunt_inductor
        """
        return self.shunt(self.delay_short(*args, **kwargs), **kwargs)

    def shunt_resistor(self, R: NumberLike, **kwargs) -> Network:
        r"""
        Shunted resistor.

        Parameters
        ----------
        R : number, array-like
            Resistor in Ohm.
        \*\*kwargs : arguments, keyword arguments
            passed to func:`resistor`

        Returns
        -------
        shunt_resistor : :class:`~skrf.network.Network` object
            shunted resistor (2-port)

        Notes
        -----
        This calls::

                shunt(resistor(R, **kwargs) ** short())

        See Also
        --------
        shunt
        shunt_delay_load
        shunt_delay_open
        shunt_delay_short
        shunt_inductor
        shunt_capacitor
        """
        return self.shunt(self.resistor(R=R, **kwargs) **
                          self.short(**kwargs), **kwargs)

    def shunt_capacitor(self, C: NumberLike, **kwargs) -> Network:
        r"""
        Shunted capacitor.

        Parameters
        ----------
        C : number, array-like
            Capacitance in Farads.
        \*\*kwargs : arguments, keyword arguments
            passed to func:`capacitor`

        Returns
        -------
        shunt_capacitor : :class:`~skrf.network.Network` object
            shunted capacitor (2-port)

        Notes
        -----
        This calls::

                shunt(capacitor(C, **kwargs) ** short())

        See Also
        --------
        shunt
        shunt_delay_load
        shunt_delay_open
        shunt_delay_short
        shunt_resistor
        shunt_inductor
        """
        return self.shunt(self.capacitor(C=C, **kwargs) **
                          self.short(**kwargs), **kwargs)

    def shunt_inductor(self, L: NumberLike, **kwargs) -> Network:
        r"""
        Shunted inductor.

        Parameters
        ----------
        L : number, array-like
            Inductance in Henries.
        \*\*kwargs : arguments, keyword arguments
            passed to func:`inductor`

        Returns
        -------
        shunt_inductor : :class:`~skrf.network.Network` object
            shunted inductor(2-port)

        Notes
        -----
        This calls::

                shunt(inductor(L, **kwargs) ** short())

        See Also
        --------
        shunt
        shunt_delay_load
        shunt_delay_open
        shunt_delay_short
        shunt_resistor
        shunt_capacitor
        """
        return self.shunt(self.inductor(L=L, **kwargs) **
                          self.short(**kwargs), **kwargs)

    def capacitor_q(self, C: NumberLike, f_0: NumberLike, q_factor: NumberLike, **kwargs) -> Network:
        r"""
        Capacitor with Q factor.

        Parameters
        ----------
        C : number, array-like
            Capacitance in Farads.
        f_0 : number
            Frequency at which Q is defined, in Hz.
        q_factor : number
            Q-factor of capacitor
        \*\*kwargs : arguments, keyword arguments
            passed to func:`capacitor`

        Returns
        -------
        capacitor_q : :class:`~skrf.network.Network` object
            capacitor_q (2-port)

        """
        idea_cap = self.shunt(self.capacitor(C=C, **kwargs), **kwargs)
        rac = q_factor / (C * 2 * np.pi * f_0)
        idea_res = self.shunt(self.resistor(R=rac), **kwargs)

        return innerconnect(connect(idea_cap, 1, idea_res, 2), 1, 3)

    def inductor_q(self,
                   L: NumberLike,
                   f_0: NumberLike,
                   q_factor: NumberLike,
                   rdc: NumberLike = 0.0,
                   **kwargs) -> Network:
        r"""
        Inductor with Q factor.

        Parameters
        ----------
        L : number, array-like
            Inductance in Henries.
        f_0 : number
            Frequency at which Q is defined, in Hz.
        q_factor : number
            Q-factor of inductor
        rdc: number, optional
            DC resistance, in Ohms. Default is 0 Ohm.
        \*\*kwargs : arguments, keyword arguments
            passed to func:`inductor`

        Returns
        -------
        inductor_q : :class:`~skrf.network.Network` object
            inductor_q (2-port)

        """
        w_q = 2 * np.pi * f_0

        if rdc == 0.0:
            rdc = 0.05 * w_q * L / q_factor

        rq1 = w_q * L / q_factor
        rq2 = np.sqrt(rq1**2 - rdc**2)
        qt = w_q * L / rq2
        rac = self.frequency.w * L / qt
        r1 = np.sqrt(rdc**2 + rac**2)

        return self.inductor(L=L, **kwargs) ** self.resistor(R=r1)

    def attenuator(self, s21: NumberLike, db: bool = True, d: Number = 0,
                   unit: str = 'deg', name: str = '', **kwargs) -> Network:
        r"""
        Ideal matched attenuator of a given length.

        Parameters
        ----------
        s21 : number, array-like
            the attenuation
        db : bool, optional
            is s21 in dB? otherwise assumes linear. Default is True (dB).
        d : number, optional
            length of attenuator. Default is 0.
        unit : ['deg','rad','m','cm','um','in','mil','s','us','ns','ps']
            the units of d.  See :func:`to_meters`, for details.
            Default is 'deg'
        name : str
            Name for the returned attenuator Network
        \*\*kwargs : arguments, keyword arguments
            passed to func:`line`

        Returns
        -------
        ntwk : :class:`~skrf.network.Network` object
            2-port attenuator

        """

        s21 = np.array(s21)
        if db:
            s21 = mf.db_2_magnitude(s21)

        result = self.match(nports=2)
        result.s[:, 0, 1] = s21
        result.s[:, 1, 0] = s21
        result = result ** self.line(d=d, unit=unit, **kwargs)
        result.name = name
        return result

    def lossless_mismatch(self, s11: NumberLike, db: bool = True, **kwargs) -> Network:
        r"""
        Lossless, symmetric mismatch defined by its return loss.

        Parameters
        ----------
        s11 : complex number, number, or array-like
            the reflection coefficient. if db==True, then phase is ignored

        db : bool, optional
            is s11 in db? otherwise assumes linear. Default is True (dB)

        \*\*kwargs : arguments, keyword arguments
            passed to func:`match`

        Returns
        -------
        ntwk : :class:`~skrf.network.Network` object
            2-port lossless mismatch

        """

        result = self.match(nports=2, **kwargs)
        s11 = np.array(s11)
        if db:
            s11 = mf.db_2_magnitude(s11)

        result.s[:, 0, 0] = s11
        result.s[:, 1, 1] = s11

        s21_mag = np.sqrt(1 - np.abs(s11) ** 2)
        s21_phase = np.angle(s11) + np.pi / 2 * (np.angle(s11) <= 0) - np.pi / 2 * (np.angle(s11) > 0)
        result.s[:, 0, 1] = s21_mag * np.exp(1j * s21_phase)
        result.s[:, 1, 0] = result.s[:, 0, 1]
        return result

    def isolator(self, source_port: int = 0, **kwargs) -> Network:
        r"""
        Two-port isolator.

        Parameters
        -------------
        source_port: int in [0,1], optional
            port at which power can flow from. Default is 0.
        \*\*kwargs : arguments, keyword arguments
            passed to func:`thru`

        Returns
        -------
        ntwk : :class:`~skrf.network.Network` object
            2-port isolator

        """
        result = self.thru(**kwargs)
        if source_port == 0:
            result.s[:, 0, 1] = 0
        elif source_port == 1:
            result.s[:, 1, 0] = 0
        return result



    ## Noisy Networks

    def white_gaussian_polar(self, phase_dev: Number, mag_dev: Number,
                             n_ports: int = 1, **kwargs) -> Network:
        r"""
        Complex zero-mean gaussian white-noise network.

        Creates a network whose s-matrix is complex zero-mean gaussian
        white-noise, of given standard deviations for phase and
        magnitude components.
        This 'noise' network can be added to networks to simulate
        additive noise.

        Parameters
        ----------
        phase_mag : number
            standard deviation of magnitude
        phase_dev : number
            standard deviation of phase
        n_ports : int
            number of ports.
        \*\*kwargs : passed to :class:`~skrf.network.Network`
            initializer

        Returns
        --------
        result : :class:`~skrf.network.Network` object
            a noise network
        """
        shape = (self.frequency.npoints, n_ports,n_ports)
        phase_rv= stats.norm(loc=0, scale=phase_dev).rvs(size = shape)
        mag_rv = stats.norm(loc=0, scale=mag_dev).rvs(size = shape)

        result = Network(**kwargs)
        result.frequency = self.frequency
        result.s = mag_rv*np.exp(1j*phase_rv)
        return result

    def random(self, n_ports: int = 1, reciprocal: bool = False, matched: bool = False,
               symmetric: bool = False, **kwargs) -> Network:
        r"""
        Complex random network.

        Creates a n-port network whose s-matrix is filled with random
        complex numbers. Optionally, result can be matched or reciprocal.

        Parameters
        ----------
        n_ports : int
            number of ports.
        reciprocal : bool
            makes s-matrix symmetric ($S_{mn} = S_{nm}$)
        symmetric : bool
            makes s-matrix diagonal have single value ($S_{mm}=S_{nn}$)
        matched : bool
            makes diagonals of s-matrix zero

        \*\*kwargs : passed to :class:`~skrf.network.Network`
                initializer

        Returns
        -------
        result : :class:`~skrf.network.Network` object
                the network
        """
        result = self.match(nports = n_ports, **kwargs)
        result.s = mf.rand_c(self.frequency.npoints, n_ports,n_ports)
        result.port_modes = np.array(["S"] * result.nports)
        if reciprocal and n_ports>1:
            for m in range(n_ports):
                for n in range(n_ports):
                    if m>n:
                        result.s[:,m,n] = result.s[:,n,m]
        if symmetric:
            for m in range(n_ports):
                for n in range(n_ports):
                    if m==n:
                        result.s[:,m,n] = result.s[:,0,0]
        if matched:
            for m in range(n_ports):
                for n in range(n_ports):
                    if m==n:
                        result.s[:,m,n] = 0

        return result

    ## OTHER METHODS
    def extract_distance(self, ntwk: Network) -> NumberLike:
        """
        Determines physical distance from a transmission or reflection Network.

        Given a matched transmission or reflection measurement the
        physical distance is estimated at each frequency point based on
        the scattering parameter phase of the ntwk and propagation constant.

        Notes
        -----
        If the Network is a reflect measurement, the returned distance will
        be twice the physical distance.

        Parameters
        ----------
        ntwk : `Network`
            A one-port network of either the reflection or the transmission.

        Returns
        -------
        d : number or array_like
            physical distance

        Examples
        --------
        >>> air = rf.air50
        >>> l = air.line(1, 'cm')
        >>> d_found = air.extract_distance(l.s21)
        >>> d_found
        """
        if ntwk.nports ==1:
            dphi = gradient(ntwk.s_rad_unwrap.flatten())
            dgamma = gradient(self.gamma.imag)
            return  -dphi/dgamma
        else:
            raise ValueError('ntwk must be one-port. Select s21 or s12 for a two-port.')



    def plot(self, *args, **kw):
        return self.frequency.plot(*args, **kw)



    def write_csv(self, filename: str = 'f,gamma,Z0,z0.csv'):
        """
        write this media's frequency, gamma, Z0, and z0 to a csv file.

        Parameters
        ----------
        filename : string, optional
            file name to write out data to.
            Default is 'f,gamma,Z0,z0.csv', so you probably want to specify it.

        See Also
        --------
        from_csv : class method to initialize Media object from a
            csv file written from this function
        """

        header = f'f[{self.frequency.unit}], Re(z0), Im(z0), Re(gamma), Im(gamma), Re(z0_port), Im(z0_port)\n'

        g,z,pz  = self.gamma, \
                self.z0, self.z0_port if self.z0_port is not None else self.z0

        data = np.vstack(\
                [self.frequency.f_scaled, z.real, z.imag, \
                g.real, g.imag, pz.real, pz.imag]).T

        np.savetxt(filename,data,delimiter=',',header=header)



class DefinedGammaZ0(Media):
    """
    A media directly defined by its propagation constant and characteristic impedance.

    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object or None
        frequency band of this transmission line medium.
        Default is None, which produces a 1-10ghz band with 101 points.
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance Z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance Z0 of the media.
        (Default is None)
    gamma : number or array-like, optional
        complex propagation constant. `gamma` must adhere to
        the following convention:

            * positive real(gamma) = attenuation
            * positive imag(gamma) = forward propagation
        Default is 1j (lossless).
    z0 : number or array-like, optional.
        complex characteristic impedance of the media.
        Default is 50 ohm.
    Z0 : number, array-like, or None
        deprecated parameter, only emmit a deprecation warning.
    """

    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0: NumberLike = 50,
                 Z0: NumberLike | None = None,
                 gamma: NumberLike = 1j):
        super().__init__(frequency=frequency, z0_port=z0_port)

        self.gamma= gamma
        self.z0_characteristic = z0

        if Z0 is not None:
            # warns of deprecation
            warnings.warn(
                'Use of `Z0` in DefinedGammaZ0 initialization is deprecated.\n'
                '`Z0` has no effect. Use `z0` instead\n'
                '`Z0` will be removed in version 1.0',
              DeprecationWarning, stacklevel = 2)

    @classmethod
    def from_csv(cls, filename: str, *args, **kwargs) -> Media:
        """
        Create a Media from numerical values stored in a csv file.

        The csv file format must be written by the function :func:`write_csv`,
        or similar method which produces the following format::

            f[$unit], Re(Z0), Im(Z0), Re(gamma), Im(gamma), Re(port Z0), Im(port Z0)
            1, 1, 1, 1, 1, 1, 1
            2, 1, 1, 1, 1, 1, 1
            .....

        See Also
        --------
        write_csv
        """
        try:
            fid = open(filename)
        except(TypeError):
            # they may have passed a file
            fid = filename

        header = fid.readline()
        # this is not the correct way to do this ... but whatever
        f_unit = header.split(',')[0].split('[')[1].split(']')[0]

        f,z_re,z_im,g_re,g_im,pz_re,pz_im = \
                np.loadtxt(fid,  delimiter=',').T

        if isinstance(filename, (str, Path)):
            fid.close()

        return cls(
            frequency = Frequency.from_f(f, unit=f_unit),
            z0 = z_re+1j*z_im,
            gamma = g_re+1j*g_im,
            z0_port = pz_re+1j*pz_im,
            **kwargs
            )

    @property
    def npoints(self):
        return self.frequency.npoints

    @npoints.setter
    def npoints(self,val):
        # this is done to trigger checks on vector lengths for
        # gamma/Z0/z0
        new_freq= self.frequency.copy()
        new_freq.npoints = val
        self.frequency = new_freq


    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, val):
        if hasattr(self, '_frequency') and self._frequency is not None:

            # they are updating the frequency, we may have to do something
            attrs_to_test = [self._gamma, self._z0, self._z0_port]
            if any([has_len(k) for k in attrs_to_test]):
                 raise NotImplementedError("updating a Media frequency, with non-constant "
                                           "gamma/Z0/z0 is not worked out yet")
        self._frequency = val

    @property
    def z0_characteristic(self):
        """
        Characteristic Impedance, :math:`z_0`.

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        return self._z0*ones(len(self))

    @z0_characteristic.setter
    def z0_characteristic(self, val):
        self._z0 = val

    @property
    def gamma(self):
        """
        Propagation constant.

        Returns
        ---------
        gamma : :class:`numpy.ndarray`
            complex propagation constant for this media

        Notes
        ------
        `gamma` must adhere to the following convention:

         * positive real(gamma) = attenuation
         * positive imag(gamma) = forward propagation
        """
        return self._gamma*ones(len(self))

    @gamma.setter
    def gamma(self, val):
        self._gamma = val

def has_len(x: NumberLike) -> bool:
    """
    Test of x has any length (ie is a vector).

    This is slightly non-trivial because [3] has len() but is
    doesn't really have any length.
    """
    try:
        return (len(array(x))>1)
    except TypeError:
        return False

def parse_z0(s: str) -> NumberLike:
    """
    Parse a z0 string.

    Parameters
    ----------
    s : str
        z0 string, like '50+10j'

    Returns
    -------
    z0 : np.ndarray

    Raises
    ------
    ValueError
        If could not arse the z0 string.
    """
    # they passed a string for z0, try to parse it
    re_numbers = re.compile(r'\d+')
    numbers = re.findall(re_numbers, s)
    if len(numbers)==2:
        out = float(numbers[0]) +1j*float(numbers[1])
    elif len(numbers)==1:
        out = float(numbers[0])
    else:
        raise ValueError('couldnt parse z0 string')
    return out



import os
import unittest

import numpy as np

import skrf as rf
from skrf.frequency import InvalidFrequencyWarning


class FrequencyTestCase(unittest.TestCase):
    """ """

    def setUp(self):
        """ """
        self.test_dir = os.path.dirname(os.path.abspath(__file__)) + "/"

    def test_create_linear_sweep(self):
        freq = rf.Frequency(1, 10, 10, "ghz")
        self.assertTrue((freq.f == np.linspace(1, 10, 10) * 1e9).all())
        self.assertTrue((freq.f_scaled == np.linspace(1, 10, 10)).all())
        self.assertTrue(freq.sweep_type == "lin")

    def test_create_log_sweep(self):
        freq = rf.Frequency(1, 10, 10, "ghz", sweep_type="log")
        # Check end points
        self.assertTrue(freq.f[0] == 1e9)
        self.assertTrue(freq.f[-1] == 10e9)
        spacing = [freq.f[i + 1] / freq.f[i] for i in range(len(freq.f) - 1)]
        # Check that frequency is increasing
        self.assertTrue(all(s > 1 for s in spacing))
        # Check that ratio of adjacent frequency points is identical
        self.assertTrue(all(abs(spacing[i] - spacing[0]) < 1e-10 for i in range(len(spacing))))
        self.assertTrue(freq.sweep_type == "log")

    def test_create_rando_sweep(self):
        f = np.array([1, 5, 200])
        freq = rf.Frequency.from_f(f, unit="khz")
        self.assertTrue((freq.f == f * 1e3).all())
        self.assertTrue((freq.f_scaled == f).all())

        # TODO : in next release
        # with self.assertRaises(AttributeError):
        #     # number of point is a property and can't be set
        #     freq.npoints = 10

    def test_rando_sweep_from_touchstone(self):
        """
        this also tests the ability to read a touchstone file.
        """
        rando_sweep_ntwk = rf.Network(os.path.join(self.test_dir, "ntwk_arbitrary_frequency.s2p"))
        self.assertTrue((rando_sweep_ntwk.f == np.array([1, 4, 10, 20])).all())

    def test_slicer(self):
        a = rf.Frequency.from_f([1, 2, 4, 5, 6], unit="GHz")

        b = a["2-5ghz"]
        tinyfloat = 1e-12
        self.assertTrue((abs(b.f - [2e9, 4e9, 5e9]) < tinyfloat).all())

    def test_frequency_check(self):
        with self.assertWarns(InvalidFrequencyWarning):
            freq = rf.Frequency.from_f([2, 1], unit="Hz")

        with self.assertWarns(InvalidFrequencyWarning):
            freq = rf.Frequency.from_f([1, 2, 2], unit="Hz")

        with self.assertWarns(InvalidFrequencyWarning):
            freq = rf.Frequency.from_f([1, 2, 2], unit="Hz")
            inv = freq.drop_non_monotonic_increasing()
            self.assertListEqual(inv, [2])
            self.assertTrue(np.allclose(freq.f, [1, 2]))

    def test_immutability(self):
        """
        To avoid corner cases, it is not be possible to change the
        frequency points directly.
        """
        a = rf.Frequency.from_f([1, 2, 4, 5, 6], unit="Hz")
        with self.assertRaises(AttributeError):
            a.f = [1, 2]
        with self.assertRaises(AttributeError):
            a.npoints = 10
        with self.assertRaises(AttributeError):
            a.start = 2
        with self.assertRaises(AttributeError):
            a.stop = 10

    def test_frequency_math(self):
        from operator import add, floordiv, mod, mul, sub, truediv

        # Test all the basic operators
        for op in (add, sub, mul, floordiv, truediv, mod):
            # Test 2 Frequency objects
            a = rf.Frequency(1, 10, 10, "GHz")
            b = rf.Frequency.from_f(10, "Hz")
            self.assertTrue(np.array_equal(op(a, b).f, op(a.f, b.f)))

            # Test a Frequency object and an integer
            c = rf.Frequency(1, 10, 10, "GHz")
            d = 5
            self.assertTrue(np.array_equal(op(c, d).f, op(c.f, d)))

            # Test a Frequency object and a float
            e = rf.Frequency(1, 10, 10, "GHz")
            f_ = 5.31
            self.assertTrue(np.array_equal(op(e, f_).f, op(e.f, f_)))

            # Test a Frequency object and a numpy array of the appropriate length
            g = rf.Frequency(1, 10, 10, "GHz")
            h = np.linspace(10, 100, g.f.size)
            self.assertTrue(np.array_equal(op(g, h).f, op(g.f, h)))

            # Test a Frequency object and a numpy array of an inappropriate length
            i = rf.Frequency(1, 10, 10, "GHz")
            j = np.linspace(10, 100, g.f.size * 2)
            with self.assertRaises(ValueError) as context:
                np.array_equal(op(i, j).f, op(i.f, j))

suite = unittest.TestLoader().loadTestsFromTestCase(FrequencyTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)



"""
.. module:: skrf.io.metas
========================================
metas (:mod:`skrf.io.metas`)
========================================

Functions for reading and writing file formats defined by METAS

https://www.metas.ch/vnatools

.. autosummary::
   :toctree: generated/


"""
from numpy import array, hstack, savetxt

from ..mathFunctions import complex2Scalar
from ..network import average
from ..networkSet import NetworkSet


def ns_2_sdatcv(ns, fname, polar=False):
    """
    Write a sdatcv from a skrf.NetworkSet.

    Parameters
    ----------
    ns : :class:`~skrf.networkset.NetworkSet`
        `NetworkSet`
    fname : string
        filename to save
    polar : boolean, optional
        Average the mag/phase components individually. Default is False.

    """

    ntwk = ns[0]
    nports = ntwk.nports
    freq_hz = ntwk.f.reshape(-1, 1)

    # make the header and columns information
    top = '\n'.join(['SDATCV',
                   'Ports',
                   '\t'.join(['%i\t'%(k+1) for k in range(nports)])])

    # port impedance info
    z0ri = complex2Scalar(ntwk.z0[0])
    zcol ='\t'.join(['Zr[%i]re\tZr[%i]im'%(k+1, k+1) \
        for k in range(nports)])
    zvals = '\t'.join([str(k) for k in z0ri])
    zhead = '\n'.join([zcol,zvals])

    # s and cov matrix info
    shead = '\t'.join(['S[%i,%i]%s'%(m+1, n+1, k) \
        for m,n in ntwk.port_tuples for k in ['re','im'] ])

    cvhead = '\t'.join(['CV[%i,%i]'%(n+1, m+1) \
        for m in range(2*nports**2) for n in range(2*nports**2)])

    datahead = '\t'.join(['Freq', shead, cvhead])

    header = '\n'.join([top, zhead, datahead])

    # calculate covariance matrix
    cv = ns.cov()

    # calculate mean s value,  everything so we have a 2D matrix
    mean_ntwk = average(ns, polar=polar)
    s_mean_flat = NetworkSet([mean_ntwk]).scalar_mat().squeeze()
    cv_flat = array([k.flatten('F') for k in cv])

    data = hstack([freq_hz, s_mean_flat, cv_flat])
    savetxt(fname, data, delimiter='\t', header=header, comments='')



from pathlib import Path

import pytest

import skrf as rf

testdir = Path(__file__).parent

@pytest.fixture()
def ntwk1() -> rf.Network:
    return rf.Network(testdir / "ntwk1.s2p")

@pytest.fixture()
def ntwk1_dc(ntwk1: rf.Network) -> rf.Network:
    return ntwk1.extrapolate_to_dc()

@pytest.fixture()
def ntwk_set_zip() -> rf.NetworkSet:
    return rf.NetworkSet.from_zip(testdir / "ntwks.zip")



"""
.. module:: skrf.vi.vna.hp
===========================================
HP VNAs (:mod:`skrf.vi.vna.hp`)
===========================================

From the age when HP made expensive test equipment rather than expensive printer ink.

"""
from .hp8510c import HP8510C



"""
data (:mod:`skrf.data`)
========================================

This Package provides data to be used in examples and testcases.

Material Properties
-------------------
.. autosummary::
   :toctree: generated/

    materials

Example/Test Networks
---------------------
.. autosummary::
   :toctree: generated/

    ring_slot
    ring_slot_meas
    line
    tee
    ind
    open_2p
    short_2p

"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from ..io.general import read
from ..network import Network

if TYPE_CHECKING:
    from ..calibration import Calibration

pwd = os.path.dirname(os.path.abspath(__file__))

class StaticData:

    @property
    def ntwk1(self) -> Network:
        return Network(os.path.join(pwd, 'ntwk1.s2p'))

    @property
    def line(self) -> Network:
        """ 2-Port Network: 'line',  75.0-110.0 GHz, 201 pts, z0=[50.+0.j 50.+0.j] """
        return Network(os.path.join(pwd, 'line.s2p'))

    @property
    def open_2p(self) -> Network:
        """ 2-Port Network: 'open',  1000000000.0-10000000000.0 Hz, 10 pts, z0=[50.+0.j 50.+0.j] """
        return Network(os.path.join(pwd, 'open.s2p'))

    @property
    def short_2p(self) -> Network:
        """ 2-Port Network: 'short',  1000000000.0-10000000000.0 Hz, 10 pts, z0=[50.+0.j 50.+0.j] """
        return Network(os.path.join(pwd, 'short.s2p'))

    @property
    def ind(self) -> Network:
        """ Measured 1 nH series inductor (with parasitics) """
        return Network(os.path.join(pwd, 'ind.s2p'))

    @property
    def ring_slot(self) -> Network:
        """ 2-Port Network: 'ring slot',  75.0-110.0 GHz, 201 pts, z0=[50.+0.j 50.+0.j] """
        return Network(os.path.join(pwd, 'ring slot.s2p'))

    @property
    def tee(self) -> Network:
        """ 3-Port Network: 'tee',  330.0-500.0 GHz, 201 pts, z0=[50.+0.j 50.+0.j 50.+0.j] """
        return Network(os.path.join(pwd, 'tee.s3p'))

    @property
    def ring_slot_meas(self) -> Network:
        """  1-Port Network: 'ring slot measured',  75.0-109.999999992 GHz, 101 pts, z0=[50.+0.j] """
        return Network(os.path.join(pwd, 'ring slot measured.s1p'))

    @property
    def wr2p2_line(self) -> Network:
        return Network(os.path.join(pwd, 'wr2p2,line.s2p'))

    @property
    def wr2p2_line1(self) -> Network:
        return Network(os.path.join(pwd, 'wr2p2,line1.s2p'))

    @property
    def wr2p2_delayshort(self) -> Network:
        return Network(os.path.join(pwd, 'wr2p2,delayshort.s1p'))

    @property
    def wr2p2_short(self) -> Network:
        return Network(os.path.join(pwd, 'wr2p2,short.s1p'))

    @property
    def wr1p5_line(self) -> Network:
        return Network(os.path.join(pwd, 'wr1p5,line.s2p'))

    @property
    def wr1p5_short(self) -> Network:
        return Network(os.path.join(pwd, 'wr1p5,short.s1p'))

    @property
    def ro_1(self) -> Network:
        return Network(os.path.join(pwd, 'ro,1.s1p'))

    @property
    def ro_2(self) -> Network:
        return Network(os.path.join(pwd, 'ro,2.s1p'))

    @property
    def ro_3(self) -> Network:
        return Network(os.path.join(pwd, 'ro,3.s1p'))

    @property
    def one_port_cal(self) -> Calibration:
        return read(os.path.join(pwd, 'one_port.cal'), encoding='latin1')

_data = StaticData()

def __getattr__(name):
    return getattr(_data, name)

#one_port_cal = read(os.path.join(pwd, 'one_port.cal'), encoding='latin1')

mpl_rc_fname = os.path.join(pwd, 'skrf.mplstyle')

# material database (taken from wikipedia)
materials = {
    'copper': {
        'resistivity(ohm*m)': 1.68e-8,
    },
    'aluminum': {
        'resistivity(ohm*m)': 2.82e-8,
    },
    'gold': {
        'resistivity(ohm*m)': 2.44e-8,
    },
    'lead': {
        'resistivity(ohm*m)': 1/4.56e6,  # from pozar appendix F
    },
    'steel(stainless)': {
        'resistivity(ohm*m)': 1/1.1e6,  # from pozar appendix F
    },
    'mylar': {
        'relative permativity': 3.1,
        'loss tangent': 500e-4,
    },
    'quartz': {
        'relative permativity': 3.8,
        'loss tangent': 1.5e-4,
    },
    'silicon': {
        'relative permativity': 11.68,
        'loss tangent': 8e-4,
    },
    'teflon': {
        'relative permativity': 2.1,
        'loss tangent': 5e-4,
    },
    'duroid 5880': {
        'relative permativity': 2.25,
        'loss tangent': 40e-4,
    },
}
"""
materials : Dictionnary of materials properties
"""

for k1,k2 in [
    ('cu', 'copper'),
    ('al', 'aluminum'),
    ('au', 'gold')]:
    materials[k1] = materials[k2]



"""
rectangularWaveguide (:mod:`skrf.media.rectangularWaveguide`)
================================================================

Represents a single mode of a homogeneously filled rectangular
waveguide of cross-section `a` x `b`. The mode is determined by
`mode-type` (`'te'` or `'tm'`) and mode indices ( `m` and `n` ).


====================================  =============  ===============
Quantity                              Symbol         Variable
====================================  =============  ===============
Characteristic Wave Number            :math:`k_0`    :attr:`k0`
Cut-off Wave Number                   :math:`k_c`    :attr:`kc`
Longitudinal Wave Number              :math:`k_z`    :attr:`gamma`
Transverse Wave Number (a)            :math:`k_x`    :attr:`kx`
Transverse Wave Number (b)            :math:`k_y`    :attr:`ky`
Characteristic Impedance              :math:`z_0`    :attr:`z0`
====================================  =============  ===============

.. autosummary::
   :toctree: generated/

   RectangularWaveguide

"""
from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING

import numpy as np
from numpy import exp, sqrt, where
from scipy.constants import epsilon_0, mu_0, pi

from ..constants import NumberLike
from ..data import materials
from ..tlineFunctions import skin_depth
from .freespace import Freespace
from .media import Media

if TYPE_CHECKING:
    from ..frequency import Frequency


class RectangularWaveguide(Media):
    r"""
    A single mode of a homogeneously filled rectangular waveguide.

    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object
        frequency band of this transmission line medium
    z0_port : number, array-like, or None
        `z0_port` is the port impedance for networks generated by the media.
        If `z0_port` is not None, the networks generated by the media are
        renormalized (or in other words embedded) from the characteristic
        impedance z0 of the media to `z0_port`.
        Else if `z0_port` is None, the networks port impedances will be the raw
        characteristic impedance z0 of the media.
        (Default is None)
    z0_override : number, array-like, or None
        `z0_override` override the characteristic impedance for the media.
        If `z0_override` is not None, the networks generated by the media have
        their characteristic impedance `z0` overrided by `z0_override`.
        (Default is None)
    z0 : number, array-like, or None
        deprecated parameter, alias to `z0_override` if `z0_override` is None.
        Emmit a deprecation warning.
    a : number, optional
        width of waveguide, in meters.
        Default is 1.
    b : number or None, optional
        height of waveguide, in meters.
        If `None` defaults to a/2.
        Default is None
    mode_type : ['te','tm']
        mode type, transverse electric (te) or transverse magnetic
        (tm) to-z. where z is direction of propagation
    m : int
        mode index in 'a'-direction
    n : int
        mode index in 'b'-direction
    ep_r : number, array-like,
        filling material's relative permittivity
    mu_r : number, array-like
        filling material's relative permeability
    rho : number, array-like, string
        resistivity (ohm-m) of the conductor walls. If array-like
        must be same length as frequency. if str, it must be a key in
        :data:`skrf.data.materials`.
    roughness : number, or array-like
        surface roughness of the conductor walls in units of RMS
        deviation from surface
    \*args, \*\*kwargs : arguments, keyword arguments
            passed to :class:`~skrf.media.media.Media`'s constructor
            (:func:`~skrf.media.media.Media.__init__`

    Examples
    --------
    Most common usage is standard aspect ratio (2:1) dominant
    mode, TE10 mode of wr10 waveguide can be constructed by

    >>> freq = rf.Frequency(75,110,101,'ghz')
    >>> rf.RectangularWaveguide(freq,a= 100*mil)
    """

    def __init__(self, frequency: Frequency | None = None,
                 z0_port: NumberLike | None = None,
                 z0_override: NumberLike | None = None,
                 z0: NumberLike | None = None,
                 a: float = 1, b: float | None = None,
                 mode_type: str = 'te', m: int = 1, n: int = 0,
                 ep_r: None | NumberLike = 1, mu_r: None | NumberLike = 1,
                 rho: None | NumberLike = None,
                 roughness: None | NumberLike = None,
                 *args, **kwargs):
        Media.__init__(self, frequency = frequency,
                       z0_port = z0_port, z0_override = z0_override, z0 = z0)

        if b is None:
            b = a/2.
        if mode_type.lower() not in ['te','tm']:
            raise ValueError('mode_type must be either \'te\' or \'tm\'')



        self.a = a
        self.b = b
        self.mode_type = mode_type.lower()
        self.m = m
        self.n = n
        self.ep_r = ep_r
        self.mu_r = mu_r
        self.rho = rho
        self.roughness = roughness


    def __str__(self):
        f=self.frequency
        output = (
                f'Rectangular Waveguide Media.  {f.f_scaled[0]}-{f.f_scaled[-1]} {f.unit}.  {f.npoints} points'
                f'\n a= {self.a:.2e}m, b= {self.b:.2e}m')
        return output

    def __repr__(self):
        return self.__str__()


    @classmethod
    def from_z0(cls, frequency: Frequency, z0: NumberLike, f: Number,
                ep_r=1, mu_r=1, **kw) -> Media:
        """
        Initialize from specified impedance at a given frequency, assuming
        the fundamental TE10 mode.

        Parameters
        ----------
        frequency : Frequency Object
        z0 : number /array
            characteristic impedance to create at `f`
        f : number
            frequency (in Hz) at which the resultant waveguide has the
            characteristic impedance z0
        ep_r : number, array-like,
            filling material's relative permittivity
        mu_r : number, array-like
            filling material's relative permeability
        """

        mu = mu_0*mu_r
        ep = epsilon_0*ep_r
        w = 2*pi*f
        a =pi/(w*mu) * 1./sqrt(1/(z0*1j)**2+ep/mu)

        kw.update(dict(frequency=frequency,a=a, m=1, n=0, ep_r=ep_r, mu_r=mu_r))

        return cls(**kw)

    @property
    def ep(self) -> NumberLike:
        r"""
        The permittivity of the filling material.

        .. math:

            \varepsilon = \varepsilon_r \varepsilon_0

        Returns
        -------
        ep : number
            filling material's permittivity in F/m.
        """
        return self.ep_r * epsilon_0

    @property
    def mu(self) -> NumberLike:
        r"""
        The permeability of the filling material.

        .. math::

            \mu = \mu_r \mu_0

        Returns
        -------
        mu : number
            filling material's permeability in H/m.

        """
        return self.mu_r * mu_0

    @property
    def k0(self) -> NumberLike:
        r"""
        Characteristic wave number.

        .. math::

            k_0 = \frac{\omega}{v} = \omega \sqrt{\varepsilon_r \mu_r}

        Returns
        -------
        k0 : number
            characteristic wave number
        """
        return 2*pi*self.frequency.f*sqrt(self.ep * self.mu)

    @property
    def ky(self) -> NumberLike:
        r"""
        Eigenvalue in the `b` direction.

        Defined as

        .. math::

            k_y = n \frac{\pi}{b}

        Returns
        -------
        ky : number
                eigenvalue in `b` direction
        """
        return self.n*pi/self.b

    @property
    def kx(self) -> NumberLike:
        r"""
        Eigenvalue in the 'a' direction.

        Defined as

        .. math::

            k_x = m \frac{\pi}{a}

        Returns
        -------
        kx : number
                eigenvalue in `a` direction
        """
        return self.m*pi/self.a

    @property
    def kc(self) -> NumberLike:
        r"""
        Cut-off wave number.

        Defined as

        .. math::

            k_c = \sqrt {k_x^2 + k_y^2} = \sqrt {
                {m \frac{\pi}{a}}^2 + {n \frac{\pi}{b}}^2}

        Returns
        -------
        kc : number
                cut-off wavenumber
        """
        return sqrt( self.kx**2 + self.ky**2)


    @property
    def f_cutoff(self) -> NumberLike:
        r"""
        cutoff frequency for this mode.

        .. math::

            f_c = \frac{v}{2 \pi} \sqrt {
                {m \frac{\pi}{a}}^2 + {n \frac{\pi}{b}}^2}

        where :math:`v= 1/\sqrt{\varepsilon \mu}`.

        """
        v = 1/sqrt(self.ep*self.mu)
        return v* self.kc/(2*np.pi)

    @property
    def f_norm(self) -> NumberLike:
        """
        Frequency vector normalized to cutoff.
        """
        return self.frequency.f/self.f_cutoff

    @property
    def rho(self) -> NumberLike:
        """
        Conductivity of sidewalls in ohm*m.

        Parameters
        ----------
        val : float, array-like or str
            the conductivity in ohm*m. If array-like must be same length
            as self.frequency. if str, it must be a key in
            :data:`skrf.data.materials`.

        Examples
        ---------
        >>> wg.rho = 2.8e-8
        >>> wg.rho = 2.8e-8 * ones(len(wg.frequency))
        >>> wg.rho = 'al'
        >>> wg.rho = 'aluminum'
        """
        if self.roughness is not None:
            delta = skin_depth(self.frequency.f, self._rho, self.mu_r)
            k_w = 1. +exp(-(delta/(2*self.roughness))**1.6)
            return self._rho*k_w**2

        return self._rho

    @rho.setter
    def rho(self, val: NumberLike | str):
        if isinstance(val, str):
            self._rho = materials[val.lower()]['resistivity(ohm*m)']
        else:
            self._rho=val

    @property
    def lambda_guide(self) -> NumberLike:
        r"""
        Guide wavelength.

        .. math::

            \lambda_g = \frac{2\pi}{\beta}

        The distance in which the phase of the field increases by 2 pi.

        See Also
        --------
        k0
        """
        return 2*pi/self.beta

    @property
    def lambda_cutoff(self) -> NumberLike:
        r"""
        Cutoff wavelength.

        .. math::

            \lambda_c = v/f_c

        where :math:`v= 1/\sqrt{\varepsilon \mu}` and :math:`f_c` the cut-off frequency.

        See Also
        --------
        f_cutoff
        """
        v = 1/sqrt(self.ep*self.mu)
        return v/self.f_cutoff

    @property
    def gamma(self) -> NumberLike:
        r"""
        The propagation constant (aka Longitudinal wave number).

        Defined as

        .. math::

            k_z = \pm j \sqrt {k_0^2 - k_c^2}

        This is:

        * IMAGINARY for propagating modes
        * REAL for non-propagating modes,

        Returns
        -------
        gamma :  number
            The propagation constant
        """
        ## haringtons form
        if False:  #self.m==1 and self.n==0:
            fs = Freespace(frequency=self.frequency,
                           ep_r=self.ep_r,
                           mu_r=self.mu_r)


            g = where(self.f_norm>1.,
                     sqrt(1-self.f_norm**(-2))*fs.gamma,  # cutton
                 -1j*sqrt(1-self.f_norm**(2))*fs.gamma)  # cutoff

        else:
            # TODO:  fix this for lossy ep/mu (remove abs?)
            k0, kc = self.k0, self.kc
            g =  1j*sqrt(abs(k0**2 - kc**2)) * (k0>kc) +\
                    sqrt(abs(kc**2- k0**2))*(k0<kc) + \
                    0*(kc==k0)

        g = g + self.alpha_c *(self.rho is not None)

        return g


    @property
    def alpha_c(self) -> NumberLike:
        r"""
        Loss due to finite conductivity and roughness of sidewalls.

        In units of np/m
        See property `rho` for setting conductivity.

        Effects of finite conductivity are taken from [#]_. If
        :attr:`roughness` is not None, then its effects the conductivity
        by


        .. math::

            \sigma_c = \frac{\sigma}{k_w^2}


        where

        .. math::

            k_w = 1 + e^{(-\delta/2h)^{1.6}}

            \delta = \mbox{skin depth}

            h = \mbox{surface roughness }


        This is taken from Ansoft HFSS help documents.

        References
        ----------

        .. [#] Chapter 9, (eq 9.8.1) of Electromagnetic Waves and Antennas by Sophocles J. Orfanidis
            http://eceweb1.rutgers.edu/~orfanidi/ewa/
        """

        if self.rho is None:
            return 0

        a,b,w,ep,rho,f_n = self.a, self.b, self.frequency.w, self.ep, \
            self.rho, self.f_norm

        return 1./b * sqrt( (w*ep)/(2./rho) ) * (1+2.*b/a*(1/f_n)**2)/\
            sqrt(1-(1/f_n)**2)

    @property
    def z0_characteristic(self) -> NumberLike:
        """
        The characteristic impedance, :math:`z_0`.

        The characteristic impedance depends of the mode ('te' or 'tm').

        Returns
        -------
        z0_characteristic : np.ndarray
            Characteristic Impedance in units of ohms
        """
        omega = self.frequency.w
        impedance_dict = {'te':   1j*omega*self.mu/(self.gamma),
                          'tm':   -1j*self.gamma/(omega*self.ep),\
                         }

        return impedance_dict[self.mode_type]



import unittest

import skrf as rf


class HomoDictTest(unittest.TestCase):
    """
    """
    def setUp(self):
        self.h = rf.util.HomoDict({'a':'asdf','b':'ZZZZ'})

    def test_get_item(self):
        self.assertEqual(self.h['a'],'asdf')


    def test_call(self):
        self.assertEqual(self.h.upper()['a'],'ASDF')


    def test_boolean_mask(self):
        match_key = [key for key in self.h.keys() if self.h[key].startswith('a') ]
        self.assertEqual(self.h[match_key], 'asdf')


class HomoListTest(unittest.TestCase):
    """
    """
    def setUp(self):
        self.h = rf.util.HomoList(['asdf','ZZZZ'])

    def test_get_item(self):
        self.assertEqual(self.h[0],'asdf')


    def test_call(self):
        self.assertEqual(self.h.upper()[0],'ASDF')


    def test_boolean_mask(self):
        match_idx = [idx for idx in range(len(self.h)) if self.h.startswith('a')]
        self.assertEqual(self.h[match_idx], 'asdf')



"""
.. currentmodule:: skrf.frequency

========================================
frequency (:mod:`skrf.frequency`)
========================================

Provides a frequency object and related functions.

Most of the functionality is provided as methods and properties of the
:class:`Frequency` Class.


Frequency Class
===============
.. autosummary::
   :toctree: generated/

   Frequency

Functions
=========

.. autosummary::
    :toctree: generated/

    overlap_freq

Misc
====

.. autosummary::
    :toctree: generated/

    InvalidFrequencyWarning

"""

from __future__ import annotations

import re
import warnings
from numbers import Number

import numpy as np
from numpy import (
    geomspace,
    gradient,  # used to center attribute `t` at 0
    linspace,
    pi,
)

from .constants import FREQ_UNITS, ZERO, FrequencyUnitT, NumberLike, SweepTypeT
from .util import Axes, axes_kwarg, find_nearest_index, slice_domain


class InvalidFrequencyWarning(UserWarning):
    """Thrown if frequency values aren't monotonously increasing
    """
    pass


class Frequency:
    """
    A frequency band.

    The frequency object provides a convenient way to work with and
    access a frequency band. It contains  a frequency vector as well as
    a frequency unit. This allows a frequency vector in a given unit
    to be available (:attr:`f_scaled`), as well as an absolute frequency
    axis in 'Hz'  (:attr:`f`).

    A Frequency object can be created from either (start, stop, npoints)
    using the default constructor, :func:`__init__`. Or, it can be
    created from an arbitrary frequency vector by using the class
    method :func:`from_f`.

    Internally, the frequency information is stored in the `f` property
    combined with the `unit` property. All other properties, `start`
    `stop`, etc are generated from these.
    """
    unit_dict = {k.lower(): k for k in FREQ_UNITS}

    """
    Dictionary to convert unit string with correct capitalization for display.
    """

    multiplier_dict={k.lower(): v for k,v in FREQ_UNITS.items()}
    """
    Frequency unit multipliers.
    """


    def __init__(self, start: float = 0, stop: float = 0, npoints: int = 0,
        unit: FrequencyUnitT | None = None, sweep_type: SweepTypeT = 'lin') -> None:
        """
        Frequency initializer.

        Creates a Frequency object from start/stop/npoints and a unit.
        Alternatively, the class method :func:`from_f` can be used to
        create a Frequency object from a frequency vector instead.

        Parameters
        ----------
        start : number, optional
            start frequency in  units of `unit`. Default is 0.
        stop : number, optional
            stop frequency in  units of `unit`. Default is 0.
        npoints : int, optional
            number of points in the band. Default is 0.
        unit : string, optional
            Frequency unit of the band: 'Hz', 'kHz', 'MHz', 'GHz', 'THz'.
            This is used to create the attribute :attr:`f_scaled`.
            It is also used by the :class:`~skrf.network.Network` class
            for plots vs. frequency. Default is 'Hz'.
        sweep_type : string, optional
            Type of the sweep: 'lin' or 'log'.
            'lin' for linear and 'log' for logarithmic. Default is 'lin'.

        Note
        ----
        The attribute `unit` sets the frequency multiplier, which is used
        to scale the frequency when `f_scaled` is referenced.

        Note
        ----
        The attribute `unit` is not case sensitive.
        Hence, for example, 'GHz' or 'ghz' is the same.

        See Also
        --------
        from_f : constructs a Frequency object from a frequency
            vector instead of start/stop/npoints.
        :attr:`unit` : frequency unit of the band

        Examples
        --------
        >>> wr1p5band = Frequency(start=500, stop=750, npoints=401, unit='ghz')
        >>> logband = Frequency(1, 1e9, 301, sweep_type='log')

        """
        if unit is None:
            warnings.warn('''
                          Frequency unit not passed: uses 'Hz' per default.
                          ''',
                          DeprecationWarning, stacklevel=2)
            unit = 'Hz'
        self._unit = unit.lower()

        start =  self.multiplier * start
        stop = self.multiplier * stop

        if npoints == 0:
            self._f = np.array([])
        elif sweep_type.lower() == 'lin':
            self._f = linspace(start, stop, npoints)
        elif sweep_type.lower() == 'log' and start > 0:
            self._f = geomspace(start, stop, npoints)
        else:
            raise ValueError('Sweep Type not recognized')

    def __str__(self) -> str:
        """
        """
        try:
            output =  \
                   '%s-%s %s, %i pts' % \
                   (self.f_scaled[0], self.f_scaled[-1], self.unit, self.npoints)
        except (IndexError):
            output = "[no freqs]"

        return output

    def __repr__(self) -> str:
        """
        """
        return self.__str__()

    def __getitem__(self, key: str | int | slice) -> Frequency:
        """
        Slices a Frequency object based on an index, or human readable string.

        Parameters
        ----------
        key : str, int, or slice
            if int, then it is interpreted as the index of the frequency
            if str, then should be like '50.1-75.5ghz', or just '50'.
            If the frequency unit is omitted then :attr:`unit` is
            used.

        Examples
        --------
        >>> b = rf.Frequency(50, 100, 101, 'ghz')
        >>> a = b['80-90ghz']
        >>> a.plot_s_db()
        """

        output = self.copy()


        if isinstance(key, str):

            # they passed a string try and do some interpretation
            re_hyphen = re.compile(r'\s*-\s*')
            re_letters = re.compile('[a-zA-Z]+')

            freq_unit = re.findall(re_letters,key)

            if len(freq_unit) == 0:
                freq_unit = self.unit
            else:
                freq_unit = freq_unit[0]

            key_nounit = re.sub(re_letters,'',key)
            edges  = re.split(re_hyphen,key_nounit)

            edges_freq = Frequency.from_f([float(k) for k in edges],
                                        unit = freq_unit)
            if len(edges_freq) ==2:
                slicer=slice_domain(output.f, edges_freq.f)
            elif len(edges_freq)==1:
                key = find_nearest_index(output.f, edges_freq.f[0])
                slicer = slice(key,key+1,1)
            else:
                raise ValueError()
            try:
                output._f = np.array(output.f[slicer]).reshape(-1)
                return output
            except(IndexError) as err:
                raise IndexError('slicing frequency is incorrect') from err


        if output.f.shape[0] > 0:
            output._f = np.array(output.f[key]).reshape(-1)
        else:
            output._f = np.empty(shape=(0))

        return output


    @classmethod
    def from_f(cls, f: NumberLike, unit: FrequencyUnitT | None = None) -> Frequency:
        """
        Construct Frequency object from a frequency vector.

        The unit is set by kwarg 'unit'

        Parameters
        ----------
        f : scalar or array-like
            frequency vector

        *args, **kwargs : arguments, keyword arguments
            passed on to  :func:`__init__`.

        Returns
        -------
        myfrequency : :class:`Frequency` object
            the Frequency object

        Raises
        ------
        InvalidFrequencyWarning:
            If frequency points are not monotonously increasing

        Examples
        --------
        >>> f = np.linspace(75,100,101)
        >>> rf.Frequency.from_f(f, unit='GHz')
        """
        if np.isscalar(f):
            f = [f]
        temp_freq =  cls(0,0,0,unit=unit)
        temp_freq._f = np.asarray(f) * temp_freq.multiplier
        temp_freq.check_monotonic_increasing()

        return temp_freq

    def __eq__(self, other: object) -> bool:
        #return (list(self.f) == list(other.f))
        # had to do this out of practicality
        if not isinstance(other, self.__class__):
            return False
        if len(self.f) != len(other.f):
            return False
        elif len(self.f) == len(other.f) == 0:
            return True
        else:
            return (max(abs(self.f-other.f)) < ZERO)

    def __ne__(self,other: object) -> bool:
        return (not self.__eq__(other))

    def __len__(self) -> int:
        """
        The number of frequency points
        """
        return self.npoints

    def __add__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f + (other.f if isinstance(other, Frequency) else other)
        return out

    def __sub__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f - (other.f if isinstance(other, Frequency) else other)
        return out

    def __mul__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f * (other.f if isinstance(other, Frequency) else other)
        return out

    def __rmul__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f * (other.f if isinstance(other, Frequency) else other)
        return out

    def __div__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f / (other.f if isinstance(other, Frequency) else other)
        return out

    def __truediv__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f / (other.f if isinstance(other, Frequency) else other)
        return out

    def __floordiv__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f // (other.f if isinstance(other, Frequency) else other)
        return out

    def __mod__(self, other: Frequency | NumberLike) -> Frequency:
        out = self.copy()
        out._f = self.f % (other.f if isinstance(other, Frequency) else other)
        return out

    def check_monotonic_increasing(self) -> None:
        """Validate the frequency values

        Raises
        ------
        InvalidFrequencyWarning:
            If frequency points are not monotonously increasing
        """
        increase = np.diff(self.f) > 0
        if not increase.all():
            warnings.warn("Frequency values are not monotonously increasing!\n"
            "To get rid of the invalid values call `drop_non_monotonic_increasing`",
                InvalidFrequencyWarning, stacklevel=2)

    def drop_non_monotonic_increasing(self) -> list[int]:
        """Drop duplicate and invalid frequency values and return the dropped indices

        Returns:
            list[int]: The dropped indices
        """
        invalid = np.diff(self.f, prepend=self.f[0]-1) <= 0
        self._f = self._f[~invalid]
        return list(np.flatnonzero(invalid))

    @property
    def start(self) -> float:
        """
        Starting frequency in Hz.
        """
        return self.f[0]

    @property
    def start_scaled(self) -> float:
        """
        Starting frequency in :attr:`unit`'s.
        """
        return self.f_scaled[0]
    @property
    def stop_scaled(self) -> float:
        """
        Stop frequency in :attr:`unit`'s.
        """
        return self.f_scaled[-1]

    @property
    def stop(self) -> float:
        """
        Stop frequency in Hz.
        """
        return self.f[-1]

    @property
    def npoints(self) -> int:
        """
        Number of points in the frequency.
        """
        return len(self.f)

    @property
    def center(self) -> float:
        """
        Center frequency in Hz.

        Returns
        -------
        center : number
            the exact center frequency in units of Hz
        """
        return self.start + (self.stop-self.start)/2.

    @property
    def center_idx(self) -> int:
        """
        Closes idx of :attr:`f` to the center frequency.
        """
        return self.npoints // 2

    @property
    def center_scaled(self) -> float:
        """
        Center frequency in :attr:`unit`'s.

        Returns
        -------
        center : number
            the exact center frequency in units of :attr:`unit`'s
        """
        return self.start_scaled + (self.stop_scaled-self.start_scaled)/2.

    @property
    def step(self) -> float:
        """
        The inter-frequency step size (in Hz) for evenly-spaced
        frequency sweeps

        See Also
        --------
        df : for general case
        """
        if self.span == 0:
            return 0.
        else:
            return self.span / (self.npoints - 1.)

    @property
    def step_scaled(self) -> float:
        """
        The inter-frequency step size (in :attr:`unit`) for evenly-spaced
        frequency sweeps.

        See Also
        --------
        df : for general case
        """
        if self.span_scaled == 0:
            return 0.
        else:
            return self.span_scaled / (self.npoints - 1.)

    @property
    def span(self) -> float:
        """
        The frequency span.
        """
        return abs(self.stop-self.start)

    @property
    def span_scaled(self) -> float:
        """
        The frequency span.
        """
        return abs(self.stop_scaled-self.start_scaled)

    @property
    def f(self) -> np.ndarray:
        """
        Frequency vector in Hz.

        Returns
        ----------
        f : :class:`numpy.ndarray`
            The frequency vector  in Hz

        See Also
        ----------
        f_scaled : frequency vector in units of :attr:`unit`
        w : angular frequency vector in rad/s
        """

        return self._f


    @property
    def f_scaled(self) -> np.ndarray:
        """
        Frequency vector in units of :attr:`unit`.

        Returns
        -------
        f_scaled : numpy.ndarray
            A frequency vector in units of :attr:`unit`

        See Also
        --------
        f : frequency vector in Hz
        w : frequency vector in rad/s
        """
        return self.f/self.multiplier

    @property
    def w(self) -> np.ndarray:
        r"""
        Angular frequency in radians/s.

        Angular frequency is defined as :math:`\omega=2\pi f` [#]_

        Returns
        -------
        w : :class:`numpy.ndarray`
            Angular frequency in rad/s

        References
        ----------
        .. [#] https://en.wikipedia.org/wiki/Angular_frequency

        See Also
        --------
        f_scaled : frequency vector in units of :attr:`unit`
        f : frequency vector in Hz
        """
        return 2*pi*self.f

    @property
    def df(self) -> np.ndarray:
        """
        The gradient of the frequency vector.

        Note
        ----
        The gradient is calculated using::

            `gradient(self.f)`

        """
        return gradient(self.f)

    @property
    def df_scaled(self) -> np.ndarray:
        """
        The gradient of the frequency vector (in unit of :attr:`unit`).

        Note
        ----
        The gradient is calculated using::

            `gradient(self.f_scaled)`
        """
        return gradient(self.f_scaled)

    @property
    def dw(self) -> np.ndarray:
        """
        The gradient of the frequency vector (in radians).

        Note
        ----
        The gradient is calculated using::

            `gradient(self.w)`
        """
        return gradient(self.w)

    @property
    def unit(self) -> FrequencyUnitT:
        """
        Unit of this frequency band.

        Possible strings for this attribute are:
        'Hz', 'kHz', 'MHz', 'GHz', 'THz'

        Setting this attribute is not case sensitive.

        Returns
        -------
        unit : string
            String representing the frequency unit
        """
        return self.unit_dict[self._unit]

    @unit.setter
    def unit(self, unit: FrequencyUnitT) -> None:
        self._unit = unit.lower()

    @property
    def multiplier(self) -> float:
        """
        Multiplier for formatting axis.

        This accesses the internal dictionary `multiplier_dict` using
        the value of :attr:`unit`

        Returns
        -------
        multiplier : number
            multiplier for this Frequencies unit
        """
        return self.multiplier_dict[self._unit]

    def copy(self) -> Frequency:
        """
        Returns a new copy of this frequency.
        """
        freq =  Frequency.from_f(self.f, unit='Hz')
        freq.unit = self.unit
        return freq

    def _t_padded(self, *, pad: int = 0, n: int | None = None, bandpass: bool | None = None) -> np.ndarray:
        if bandpass is None:
            bandpass = self.f[0] != 0

        if n is None:
            n = self.npoints + pad
            n = n if bandpass else n * 2 - 1

        if bandpass:
            dt = 1 / (n * self.step)
            t_stop = (n - 1) // 2 * dt
            t_start = -t_stop - dt if n % 2 == 0 else (-n // 2 + 1) * dt

            t = np.linspace(t_start, t_stop, num=n, endpoint=True)
        else:
            dt = 1 / (n * self.step)
            t = np.linspace(-dt * (n // 2), dt * (n // 2), num=n, endpoint=True)

        return t

    @property
    def t(self) -> np.ndarray:
        """
        Time vector in s.

        t_period = 2*(n-1)/f_step
        """
        return self._t_padded(bandpass=True)

    @property
    def t_ns(self) -> np.ndarray:
        """
        Time vector in ns.

        t_period = 2*(n-1)/f_step
        """
        return self.t*1e9

    def round_to(self, val: FrequencyUnitT | Number = 'Hz') -> None:
        """
        Round off frequency values to a specified precision.

        This is useful for dealing with finite precision limitations of
        VNA's and/or other software

        Parameters
        ----------
        val : string or number
            if val is a string it should be a frequency :attr:`unit`
            (ie 'Hz', 'MHz',etc). if its a number, then this returns
            f = f-f%val

        Examples
        --------
        >>> f = skrf.Frequency.from_f([.1,1.2,3.5],unit='Hz')
        >>> f.round_to('Hz')

        """
        if isinstance(val, str):
            val = self.multiplier_dict[val.lower()]

        self.f = np.round(self.f/val)*val

    def overlap(self,f2: Frequency) -> Frequency:
        """
        Calculates overlapping frequency  between self and f2.

        See Also
        --------
        overlap_freq

        """
        return overlap_freq(self, f2)

    @property
    def sweep_type(self) -> str:
        """
        Frequency sweep type.

        Returns
        -------
        sweep_type: str
            'lin' if linearly increasing, 'log' or 'unknown'.

        """
        if np.allclose(self.f, linspace(self.f[0], self.f[-1], self.npoints), rtol=0.05):
            sweep_type = 'lin'
        elif self.f[0] and np.allclose(self.f, geomspace(self.f[0], self.f[-1], self.npoints), rtol=0.05):
            sweep_type = 'log'
        else:
            sweep_type = 'unknown'
        return sweep_type

    @axes_kwarg
    def labelXAxis(self, ax: Axes = None):
        """
        Label the x-axis of a plot.

        Sets the labels of a plot using :func:`matplotlib.x_label` with
        string containing the frequency unit.

        Parameters
        ----------
        ax : :class:`matplotlib.Axes` or None, optional
                Axes on which to label the plot.
                Defaults is None, for the current axe
                returned by :func:`matplotlib.gca()`
        """

        ax.set_xlabel(f'Frequency ({self.unit})')

    @axes_kwarg
    def plot(self, y: NumberLike, *args, ax: Axes=None, **kwargs):
        """
        Plot something vs this frequency.

        This plots whatever is given vs. `self.f_scaled` and then
        calls `labelXAxis`.
        """

        from .plotting import scale_frequency_ticks

        try:
            if len(np.shape(y)) > 2:
                # perhaps the dimensions are empty, try to squeeze it down
                y = y.squeeze()
                if len(np.shape(y)) > 2:
                    # the dimensions are full, so lets loop and plot each
                    for m in range(np.shape(y)[1]):
                        for n in range(np.shape(y)[2]):
                            self.plot(y[:, m, n], *args, **kwargs)
                    return
            if len(y) == len(self):
                pass
            else:

                raise IndexError(['thing to plot doesn\'t have same'
                                ' number of points as f'])
        except(TypeError):
            y = y * np.ones(len(self))

        # plt.plot(self.f_scaled, y, *args, **kwargs)
        ax.plot(self.f, y, *args, **kwargs)
        scale_frequency_ticks(ax, self.unit)
        ax.autoscale(axis='x', tight=True)
        self.labelXAxis()


def overlap_freq(f1: Frequency,f2: Frequency) -> Frequency:
    """
    Calculates overlapping frequency between f1 and f2.

    Or, put more accurately, this returns a Frequency that is the part
    of f1 that is overlapped by f2. The resultant start frequency is
    the smallest f1.f that is greater than f2.f.start, and the stop
    frequency is the largest f1.f that is smaller than f2.f.stop.

    This way the new frequency overlays onto f1.


    Parameters
    ----------
    f1 : :class:`Frequency`
        a frequency object
    f2 : :class:`Frequency`
        a frequency object

    Returns
    -------
    f3 : :class:`Frequency`
        part of f1 that is overlapped by f2

    """
    if f1.start > f2.stop:
        raise ValueError('Out of bounds. f1.start > f2.stop')
    elif f2.start > f1.stop:
        raise ValueError('Out of bounds. f2.start > f1.stop')


    start = max(f1.start, f2.start)
    stop = min(f1.stop, f2.stop)
    f = f1.f[(f1.f>=start) & (f1.f<=stop)]
    freq =  Frequency.from_f(f, unit = 'Hz')
    freq.unit = f1.unit
    return freq



"""
.. module:: skrf.io.touchstone

========================================
touchstone (:mod:`skrf.io.touchstone`)
========================================

Touchstone class and utilities

.. autosummary::
   :toctree: generated/

   Touchstone

Functions related to reading/writing touchstones.
-------------------------------------------------

.. autosummary::
   :toctree: generated/

   hfss_touchstone_2_gamma_z0
   hfss_touchstone_2_media
   hfss_touchstone_2_network
   read_zipped_touchstones

"""
from __future__ import annotations

import os
import re
import typing
import warnings
import zipfile
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Callable, Literal

import numpy as np

from ..constants import FREQ_UNITS, S_DEF_HFSS_DEFAULT, S_DEFINITIONS, SparamFormatT
from ..media import DefinedGammaZ0
from ..network import Network
from ..util import get_fid


def remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


@dataclass
class ParserState:
    """Class to hold dynamic variables while parsing the touchstone file.
    """
    rank: int | None = None
    option_line_parsed: bool = False
    hfss_gamma: list[list[float]] = field(default_factory=list)
    hfss_impedance: list[list[float]] = field(default_factory=list)
    comments: list[str] = field(default_factory=list)
    comments_after_option_line: list[str] = field(default_factory=list)
    matrix_format: str = "full"
    parse_network: bool = True
    _parse_noise: bool = False
    f: list[float] = field(default_factory=list)
    f_noise: list[float] = field(default_factory=list)
    s: list[float] = field(default_factory=list)
    noise: list[float] = field(default_factory=list)
    two_port_order_legacy: bool = True
    number_noise_freq: int = 0
    port_names: dict[int, str] = field(default_factory=dict)
    ansys_data_type: str | None = None
    mixed_mode_order: list[str] | None = None
    frequency_unit: str = "ghz"
    parameter: str = "s"
    format: str = "ma"
    resistance: complex = complex(50)

    @property
    def n_ansys_impedance_values(self) -> int:
        """Returns the number of port impedances returned by Ansys HFSS.

        Currently this function returns rank * 2.

        Returns:
            int: number of impedance values.
        """
        # See https://github.com/scikit-rf/scikit-rf/issues/354 for details.
        #if self.ansys_data_type == "terminal":
        #    return self.rank**2 * 2

        return self.rank * 2

    @cached_property
    def numbers_per_line(self) -> int:
        """Returns data points per frequency point.

        Returns:
            int: Number of data points per frequency point.
        """
        if self.matrix_format == "full":
            return self.rank**2 * 2
        return self.rank*(self.rank+1)

    @property
    def parse_noise(self) -> bool:
        """Returns true if the parser expects noise data.

        Returns:
            bool: True, if noise data is expected.
        """
        return self._parse_noise

    @parse_noise.setter
    def parse_noise(self, x: bool) -> None:
        self.parse_network = False
        self._parse_noise = x

    @property
    def frequency_mult(self) -> float:
        _units = {k.lower(): v for k,v in FREQ_UNITS.items()}
        return _units[self.frequency_unit]

    def parse_port(self, line: str):
        """Regex parser for port names.

        Args:
            line (str): Touchstone line.
        """
        m = re.match(r"! Port\[(\d+)\]\s*=\s*(.*)$", line)
        if m:
            self.port_names[int(m.group(1)) - 1] = m.group(2)

    def append_comment(self, line: str) -> None:
        """Append comment, and append appropriate comment list.

        Args:
            line (str): Line to parse
        """
        if self.option_line_parsed:
            self.comments_after_option_line.append(line)
        else:
            self.comments.append(line)

    def parse_option_line(self, line: str) -> None:
        """Parse the option line starting with #

        Args:
            line (str): Line to parse

        Raises:
            ValueError: If option line contains invalid options.

        """
        if self.option_line_parsed:
            return
        toks = line.lower()[1:].strip().split()
        # fill the option line with the missing defaults
        toks.extend(["ghz", "s", "ma", "r", "50"][len(toks) :])
        self.frequency_unit = toks[0]
        self.parameter = toks[1]
        self.format = toks[2]
        self.resistance = complex(toks[4])
        err_msg = ""
        if self.frequency_unit not in ["hz", "khz", "mhz", "ghz"]:
            err_msg = f"ERROR: illegal frequency_unit {self.frequency_unit}\n"
        if self.parameter not in "syzgh":
            err_msg = f"ERROR: illegal parameter value {self.parameter}\n"
        if self.format not in typing.get_args(SparamFormatT):
            err_msg = f"ERROR: illegal format value {self.format}\n"

        if err_msg:
            raise ValueError(err_msg)

        self.option_line_parsed = True


class Touchstone:
    """
    Class to read touchstone s-parameter files.

    The reference for writing this class is the draft of the
    Touchstone(R) File Format Specification Rev 2.0 [#]_ and
    Touchstone(R) File Format Specification Version 2.0 [#]_

    References
    ----------
    .. [#] https://ibis.org/interconnect_wip/touchstone_spec2_draft.pdf
    .. [#] https://ibis.org/touchstone_ver2.0/touchstone_ver2_0.pdf
    """

    def __init__(self, file: str | Path | typing.TextIO, encoding: str | None = None):
        """
        constructor

        Parameters
        ----------
        file : str, Path, or file-object
            touchstone file to load
        encoding : str, optional
            define the file encoding to use. Default value is None,
            meaning the encoding is guessed (ANSI, UTF-8 or Latin-1).

        Examples
        --------
        From filename

        >>> t = rf.Touchstone('network.s2p')

        File encoding can be specified to help parsing the special characters:

        >>> t = rf.Touchstone('network.s2p', encoding='ISO-8859-1')

        From file-object

        >>> file = open('network.s2p')
        >>> t = rf.Touchstone(file)

        From a io.StringIO object

        >>> link = 'https://raw.githubusercontent.com/scikit-rf/scikit-rf/master/examples/
            basic_touchstone_plotting/horn antenna.s1p'
        >>> r = requests.get(link)
        >>> stringio = io.StringIO(r.text)
        >>> stringio.name = 'horn.s1p'  # must be provided for the Touchstone parser
        >>> horn = rf.Touchstone(stringio)

        """
        ## file format version.
        # Defined by default to 1.0, since version number can be omitted in V1.0 format
        self._version = "1.0"
        ## comments in the file header
        self.comments = ""
        ## unit of the frequency (Hz, kHz, MHz, GHz)
        self.frequency_unit = None
        ## number of frequency points
        self.frequency_nb = None
        ## s-parameter type (S,Y,Z,G,H)
        self.parameter = None
        ## s-parameter format (MA, DB, RI)
        self.format = None
        ## reference resistance, global setup
        self.resistance = None
        ## reference impedance for each s-parameter
        self.reference = None

        ## numpy array of original noise data
        self.noise = None

        ## kind of s-parameter data (s1p, s2p, s3p, s4p)
        self.rank = None
        ## Store port names in a list if they exist in the file
        self.port_names = None

        self.comment_variables = None
        # Does the input file has HFSS per frequency port impedances
        self.has_hfss_port_impedances = False
        self.gamma = None
        self.z0 = None
        self.s_def = None
        self.port_modes = np.array([])

        # open the file depending on encoding
        # Guessing the encoding by trial-and-error, unless specified encoding
        try:
            try:
                if encoding is not None:
                    fid = get_fid(file, encoding=encoding)
                    self.filename = fid.name
                    self.load_file(fid)
                else:
                    # Assume default encoding
                    fid = get_fid(file)
                    self.filename = fid.name
                    self.load_file(fid)
            except Exception as e:
                fid.close()
                raise e

        except UnicodeDecodeError:
            # Unicode fails -> Force Latin-1
            fid = get_fid(file, encoding="ISO-8859-1")
            self.filename = fid.name
            self.load_file(fid)

        except ValueError:
            # Assume Microsoft UTF-8 variant encoding with BOM
            fid = get_fid(file, encoding="utf-8-sig")
            self.filename = fid.name
            self.load_file(fid)

        except Exception as e:
            raise ValueError("Something went wrong by the file opening") from e

        finally:
            fid.close()

    @staticmethod
    def _parse_n_floats(*, line: str, fid: typing.TextIO, n: int, before_comment: bool) -> list[float]:
        """Parse a specified number of floats either in our outside a comment.

        Args:
            line (str): Actual line to parse.
            fid (typing.TextIO): File descriptor to get new lines if necessary.
            n (int): Number of floats to parse.
            before_comment (bool): True, if the floats should get parsed in or outside a comment.

        Returns:
            list[float]: The parsed float values
        """
        def get_part_of_line(line: str) -> str:
            """Return either the part after or before a exclamation mark.

            Args:
                line (str): Line to parse.

            Returns:
                str: string subset.
            """
            if before_comment:
                return line.partition("!")[0]
            else:
                return line.rpartition("!")[2]

        values = get_part_of_line(line).split()
        ret = []
        while True:
            if len(ret) == n:
                break

            if not values:
                values = get_part_of_line(fid.readline()).split()
            try:
                ret.append(float(values.pop(0)))
            except ValueError:
                pass

        return ret

    @property
    def version(self) -> str:
        """The version string.

        Returns:
            str: Version
        """
        return self._version

    @version.setter
    def version(self, x: str) -> None:
        self._version = x
        if x == "2.0":
            self._parse_dict.update(self._parse_dict_v2)

    def _parse_file(self, fid: typing.TextIO) -> ParserState:
        """
        Parse the raw file and generate an structured view.

        Parameters
        ----------
        fid : file object

        Returns
        -------
        state: File content as ParserState

        """
        state = ParserState()

        # Check the filename extension.
        # Should be .sNp for Touchstone format V1.0, and .ts for V2
        extension = self.filename.split(".")[-1].lower()

        m = re.match(r"[ghsyz](\d+)p", extension)
        if m:
            state.rank = int(m.group(1))
        elif extension != "ts":
            msg = (f"{self.filename} does not have a s-parameter extension ({extension})."
                    "Please, correct the extension to of form: 'sNp', where N is any integer for Touchstone v1,"
                    "or ts for Touchstone v2.")
            raise ValueError(msg)

        # Lookup dictionary for parser
        # Dictionary has string keys and values contains functions which
        # need the current line as string argument. The type hints allow
        # the IDE to provide full typing support
        # Take care of the order of the elements when inserting new key words.
        self._parse_dict: dict[str, Callable[[str], None]] = {
            "[version]": lambda x: setattr(self, "version", x.split()[1]),
            "#": lambda x: state.parse_option_line(x),
            "! gamma": lambda x: state.hfss_gamma.append(
                self._parse_n_floats(line=x, fid=fid, n=state.rank * 2, before_comment=False)
            ),
            "! port impedance": lambda x: state.hfss_impedance.append(
                self._parse_n_floats(
                    line=remove_prefix(x.lower(), "! port impedance"),
                    fid=fid,
                    n=state.n_ansys_impedance_values,
                    before_comment=False,
                )
            ),
            "! port": state.parse_port,
            "! terminal data exported": lambda _: setattr(state, "ansys_data_type", "terminal"),
            "! modal data exported": lambda _: setattr(state, "ansys_data_type", "modal"),
            "!": state.append_comment,
        }

        self._parse_dict_v2: dict[str, Callable[[str], None]] = {
            "[number of ports]": lambda x: setattr(state, "rank", int(x.split()[3])),
            "[reference]": lambda x: setattr(
                state, "resistance", self._parse_n_floats(line=x, fid=fid, n=state.rank, before_comment=True)
            ),
            "[number of frequencies]": lambda x: setattr(self, "frequency_nb", int(x.split()[3])),
            "[matrix format]": lambda x: setattr(state, "matrix_format", x.split()[2].lower()),
            "[network data]": lambda _: setattr(state, "parse_network", True),
            "[noise data]": lambda _: setattr(state, "parse_noise", True),
            "[two-port data order]": lambda x: setattr(state, "two_port_order_legacy", "21_12" in x),
            "[number of noise frequencies]": lambda x: setattr(
                state, "number_noise_freq", int(x.partition("]")[2].strip())
            ),
            "[mixed-mode order]": lambda line: setattr(state, "mixed_mode_order", line.lower().split()[2:]),
            "[end]": lambda x: None,
        }

        while True:
            line = fid.readline()
            if not line:
                break

            line_l = line.lower()

            is_data_line = True
            # Avoid traversing the self._parse_dict for each line by checking the first letter
            # {"!", "#", "["} covers all the first letters of the key of the current self._parse_dict
            if line_l[0] in {"!", "#", "["}:
                for k, v in self._parse_dict.items():
                    if line_l.startswith(k):
                        v(line)
                        is_data_line = False
                        break
            if is_data_line:
                if "!" in line:
                    line = line.partition("!")[0]
                values = list(map(float, line.split()))
                if not values:
                    continue

                if (
                    state.f
                    and len(state.s) % state.numbers_per_line == 0
                    and values[0] < state.f[-1]
                    and state.rank == 2
                    and self.version == "1.0"
                ):
                    state.parse_noise = True

                if state.parse_network:
                    if len(state.s) % state.numbers_per_line == 0:
                        state.f.append(values.pop(0))
                    state.s.extend(values)

                elif state.parse_noise:
                    state.noise.append(values)

        return state

    def load_file(self, fid: typing.TextIO):
        """
        Load the touchstone file into the internal data structures.

        Parameters
        ----------
        fid : file object

        """

        state = self._parse_file(fid=fid)

        self.comments = "\n".join([line.strip()[1:] for line in state.comments])
        self.comments_after_option_line = "\n".join([line.strip()[1:] for line in state.comments_after_option_line])
        self.rank = state.rank
        self.frequency_unit = state.frequency_unit
        self.parameter = state.parameter
        self.format = state.format
        self.resistance = state.resistance

        if state.port_names:
            self.port_names = [""] * state.rank
            for k, v in state.port_names.items():
                self.port_names[k] = v

        if state.hfss_gamma:
            self.gamma = np.array(state.hfss_gamma).view(np.complex128)


        # Impedance is parsed in the following order:
        # - HFSS comments for each frequency point and each port.
        # - TS v2 Reference keyword for each port.
        # - Reference impedance from option line.
        if state.hfss_impedance:
            self.z0 = np.array(state.hfss_impedance).view(np.complex128)
            # Comment the line in, when we need when to expect port impedances in NxN format.
            # See https://github.com/scikit-rf/scikit-rf/issues/354 for details.
            #if state.ansys_data_type == "terminal":
            #    self.z0 = np.diagonal(self.z0.reshape(-1, self.rank, self.rank), axis1=1, axis2=2)

            self.s_def = S_DEF_HFSS_DEFAULT
            self.has_hfss_port_impedances = True
            # Load the reference impedance convention from the comments
            for s_def in S_DEFINITIONS:
                if f'S-parameter uses the {s_def} definition' in self.comments:
                    self.s_def = s_def
        elif self.reference is None:
            self.z0 = np.broadcast_to(self.resistance, (len(state.f), state.rank)).copy()
        else:
            self.z0 = np.empty((len(state.f), state.rank), dtype=complex).fill(self.reference)

        self.f = np.array(state.f)
        if not len(self.f):
            self.s = np.empty((0, state.rank, state.rank))
            return

        raw = np.array(state.s).reshape(len(self.f), -1)

        if self.format == "db":
            raw[:, 0::2] = 10 ** (raw[:, 0::2] / 20.0)

        if self.format in (["ma", "db"]):
            s_flat = raw[:, 0::2] * np.exp(1j * raw[:, 1::2] * np.pi / 180)
        elif self.format == "ri":
            s_flat = raw.view(np.complex128)

        self.s_flat = s_flat

        self.s = np.empty((len(self.f), state.rank * state.rank), dtype=complex)
        if state.matrix_format == "full":
            self.s[:] = s_flat
        else:
            index = np.tril_indices(state.rank) if state.matrix_format == "lower" else np.triu_indices(state.rank)
            index_flat = np.ravel_multi_index(index, (state.rank, state.rank))
            self.s[:, index_flat] = s_flat

        if state.rank == 2 and state.two_port_order_legacy:
            self.s = np.transpose(self.s.reshape((-1, state.rank, state.rank)), axes=(0, 2, 1))
        else:
            self.s = self.s.reshape((-1, state.rank, state.rank))

        if state.matrix_format == "upper":
            index_lower = np.tril_indices(state.rank)
            self.s[(...,*index_lower)] = self.s.transpose(0, 2, 1)[(...,*index_lower)]
        elif state.matrix_format == "lower":
            index_upper = np.triu_indices(state.rank)
            self.s[(...,*index_upper)] = self.s.transpose(0, 2, 1)[(...,*index_upper)]

        self.port_modes = np.array(["S"] * state.rank)
        if state.mixed_mode_order:
            new_order = [None] * state.rank
            for i, mm in enumerate(state.mixed_mode_order):
                if mm.startswith("s"):
                    new_order[i] = int(mm[1:]) - 1
                else:
                    p1, p2 = sorted([int(e) - 1 for e in mm[1:].split(",")])

                    if mm.startswith("d"):
                        new_order[i] = p1
                    else:
                        new_order[i] = p2
                self.port_modes[new_order[i]] = mm[0].upper()

            order = np.arange(self.rank, dtype=int)
            self.s[:, new_order, :] = self.s[:, order, :]
            self.s[:, :, new_order] = self.s[:, :, order]
            self.z0[:, self.port_modes == "D"] *= 2
            self.z0[:, self.port_modes == "C"] /= 2

        if self.parameter in ["g", "h", "y", "z"]:
            if self.version == "1.0":
                self.s = self.s * self.z0[:,:, None]

            func_name = f"{self.parameter}2s"
            from .. import network
            self.s: np.ndarray = getattr(network, func_name)(self.s, self.z0)


        # multiplier from the frequency unit
        self.frequency_mult = state.frequency_mult

        if state.noise:
            self.noise = np.array(state.noise)
            self.noise[:, 0] *= self.frequency_mult

        self.f *= self.frequency_mult

    @property
    def sparameters(self) -> np.ndarray:
        """Touchstone data in tabular format.

        Returns:
            np.ndarray: Frequency and data array.
        """
        warnings.warn("This method is deprecated and will be removed.", DeprecationWarning, stacklevel=2)
        return np.hstack((self.f[:, None], self.s_flat.view(np.float64).reshape(len(self.f), -1)))

    def get_comments(self, ignored_comments: list[str]=None) -> str:
        """
        Returns the comments which appear anywhere in the file.

        Comment lines containing ignored comments are removed.
        By default these are comments which contain special meaning withing
        skrf and are not user comments.

        Returns
        -------
        processed_comments : string

        """
        if ignored_comments is None:
            ignored_comments = ["Created with skrf"]
        processed_comments = ""
        if self.comments is None:
            self.comments = ""
        for comment_line in self.comments.split("\n"):
            for ignored_comment in ignored_comments:
                if ignored_comment in comment_line:
                    comment_line = None
            if comment_line:
                processed_comments = processed_comments + comment_line + "\n"
        return processed_comments

    def get_comment_variables(self) -> dict[str, str]:
        """
        Convert hfss variable comments to a dict of vars.

        Returns
        -------
        var_dict : dict (numbers, units)
            Dictionnary containing the comments
        """
        comments = self.comments
        p1 = re.compile(r"\w* = \w*.*")
        p2 = re.compile(r"\s*(\d*\.?\d*)\s*(\w*)")
        var_dict = {}
        for k in re.findall(p1, comments):
            try:
                var, value = k.split("=")
                var = var.rstrip()
                var_dict[var] = p2.match(value).groups()
            except ValueError:
                pass
        return var_dict

    def get_format(self, format: Literal[SparamFormatT, Literal["orig"]]="ri") -> str:
        """
        Returns the file format string used for the given format.

        This is useful to get some information.

        Returns
        -------
        format : string

        """
        if format == "orig":
            frequency = self.frequency_unit
            format = self.format
        else:
            frequency = "hz"
        return f"{frequency} {self.parameter} {format.upper()} r {self.resistance}"

    def get_sparameter_names(self, format: SparamFormatT = "ri") -> list[str]:
        """
        Generate a list of column names for the s-parameter data.
        The names are different for each format.

        Parameters
        ----------
        format : str
          Format: ri, ma, db

        Returns
        -------
        names : list
            list of strings

        """
        warnings.warn("This method is deprecated and will be removed.", DeprecationWarning, stacklevel=2)
        return self.get_sparameter_data(format).keys()

    def get_sparameter_data(self, format: SparamFormatT = "ri") -> dict[str, np.ndarray]:
        """
        Get the data of the s-parameter with the given format.

        Parameters
        ----------
        format : str
          Format: ri, ma, db

        supported formats are:
          ri:    data in real/imaginary
          ma:    data in magnitude and angle (degree)
          db:    data in log magnitude and angle (degree)

        Returns
        -------
        ret: list
            list of numpy.arrays

        """
        warnings.warn("This method is deprecated and will be removed.", DeprecationWarning, stacklevel=2)
        ret = {"frequency": self.f}
        for j in range(self.rank):
            for k in range(self.rank):
                prefix = f"S{j+1}{k+1}"
                val = self.s[:, j, k]
                if self.rank == 2 and self.filename.split(".")[-1].lower() == "s2p":
                    prefix = f"S{k+1}{j+1}"
                    val = self.s[:, k, j]

                if format == "ri":
                    ret[f"{prefix}R"] = val.real
                    ret[f"{prefix}I"] = val.imag
                elif format == "ma":
                    ret[f"{prefix}M"] = np.abs(val)
                    ret[f"{prefix}A"] = np.angle(val, deg=True)
                elif format == "db":
                    ret[f"{prefix}DB"] = 20 * np.log10(np.abs(val))
                    ret[f"{prefix}A"] = np.angle(val, deg=True)

        return ret

    def get_sparameter_arrays(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns the s-parameters as a tuple of arrays.

        The first element is the frequency vector (in Hz) and the s-parameters are a 3d numpy array.
        The values of the s-parameters are complex number.

        Returns
        -------
        param : tuple of arrays

        Examples
        --------
        >>> f, a = self.sgetparameter_arrays()
        >>> s11 = a[:, 0, 0]

        """
        return self.f, self.s

    def get_noise_names(self):
        raise NotImplementedError("not yet implemented")

    def get_noise_data(self):
        # TBD = 1
        # noise_frequency = noise_values[:,0]
        # noise_minimum_figure = noise_values[:,1]
        # noise_source_reflection = noise_values[:,2]
        # noise_source_phase = noise_values[:,3]
        # noise_normalized_resistance = noise_values[:,4]
        raise NotImplementedError("not yet implemented")

    def get_gamma_z0(self):
        """
        Extracts Z0 and Gamma comments from touchstone file (if provided).

        Returns
        -------
        gamma : complex np.ndarray
            complex  propagation constant
        z0 : np.ndarray
            complex port impedance
        """
        return self.gamma, self.z0


def hfss_touchstone_2_gamma_z0(filename: str) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Extracts Z0 and Gamma comments from touchstone file.

    Takes a HFSS-style Touchstone file with Gamma and Z0 comments and
    extracts a triplet of arrays being: (frequency, Gamma, Z0)

    Parameters
    ----------
    filename : string
        the HFSS-style Touchstone file


    Returns
    -------
    f : np.ndarray
        frequency vector (in Hz)
    gamma : complex np.ndarray
        complex  propagation constant
    z0 : np.ndarray
        complex port impedance

    Examples
    --------
    >>> f,gamm,z0 = rf.hfss_touchstone_2_gamma_z0('line.s2p')
    """
    ntwk = Network(filename)

    return ntwk.frequency.f, ntwk.gamma, ntwk.z0


def hfss_touchstone_2_media(filename: str) -> list[DefinedGammaZ0]:
    """
    Creates a :class:`~skrf.media.Media` object from a a HFSS-style Touchstone file with Gamma and Z0 comments.

    Parameters
    ----------
    filename : string
        the HFSS-style Touchstone file

    Returns
    -------
    my_media : :class:`~skrf.media.media.Media` object
        the transmission line model defined by the gamma, and z0
        comments in the HFSS file.

    Examples
    --------
    >>> port1_media, port2_media = rf.hfss_touchstone_2_media('line.s2p')

    See Also
    --------
    hfss_touchstone_2_gamma_z0 : returns gamma, and z0
    """
    ntwk = Network(filename)

    freq = ntwk.frequency
    gamma = ntwk.gamma
    z0 = ntwk.z0

    media_list = []

    for port_n in range(gamma.shape[1]):
        media_list.append(DefinedGammaZ0(frequency=freq, gamma=gamma[:, port_n], z0=z0[:, port_n]))

    return media_list


def hfss_touchstone_2_network(filename: str) -> Network:
    """
    Creates a :class:`~skrf.Network` object from a a HFSS-style Touchstone file.

    Parameters
    ----------
    filename : string
        the HFSS-style Touchstone file

    Returns
    -------
    my_network : :class:`~skrf.network.Network` object
        the n-port network model

    Examples
    --------
    >>> my_network = rf.hfss_touchstone_2_network('DUT.s2p')

    See Also
    --------
    hfss_touchstone_2_gamma_z0 : returns gamma, and z0
    """
    my_network = Network(file=filename)
    return my_network


def read_zipped_touchstones(ziparchive: zipfile.ZipFile, dir: str = "") -> dict[str, Network]:
    """
    similar to skrf.io.read_all_networks, which works for directories but only for Touchstones in ziparchives.

    Parameters
    ----------
    ziparchive : :class:`zipfile.ZipFile`
        an zip archive file, containing Touchstone files and open for reading
    dir : str
        the directory of the ziparchive to read networks from, default is "" which reads only the root directory

    Returns
    -------
    dict
        keys are touchstone filenames without extensions
        values are network objects created from the touchstone files
    """
    networks = dict()
    for fname in ziparchive.namelist():  # type: str
        directory = os.path.split(fname)[0]
        if dir == directory and  re.search(r"s\d+p$", fname.lower()):
            network = Network.zipped_touchstone(fname, ziparchive)
            networks[network.name] = network
    return networks
